@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
+static char rcsid[] = "$Id: os_unix.c,v 4.101 1996/03/17 23:00:08 mikes Exp $";
 #endif
 /*
  * Program:	Operating system dependent routines - Ultrix 4.1
@@ -15,28 +15,16 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
  *
  * Please address all bugs and comments to "pine-bugs@cac.washington.edu"
  *
- * Copyright 1991-1994  University of Washington
  *
- *  Permission to use, copy, modify, and distribute this software and its
- * documentation for any purpose and without fee to the University of
- * Washington is hereby granted, provided that the above copyright notice
- * appears in all copies and that both the above copyright notice and this
- * permission notice appear in supporting documentation, and that the name
- * of the University of Washington not be used in advertising or publicity
- * pertaining to distribution of the software without specific, written
- * prior permission.  This software is made available "as is", and
- * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
- * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
- * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
- * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
- * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
- * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- * Pine and Pico are trademarks of the University of Washington.
- * No commercial use of these trademarks may be made without prior
- * written permission of the University of Washington.
+ * Pine and Pico are registered trademarks of the University of Washington.
+ * No commercial use of these trademarks may be made without prior written
+ * permission of the University of Washington.
+ * 
+ * Pine, Pico, and Pilot software and its included text are Copyright
+ * 1989-1996 by the University of Washington.
+ * 
+ * The full text of our legal notices is contained in the file called
+ * CPYRIGHT, included with this distribution.
  *
  *
  * Notes:
@@ -50,6 +38,9 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
  * - 3B2, 3b1/7300, SCO ports by:
  *       rll@felton.felton.ca.us, 7 Feb. 1993
  *
+ * - Altos System V (asv) port by:
+ *	 Tim Rice <tim@trr.metro.net>    6 Mar 96
+ *
  * - Probably have to break this up into separate os_type.c files since
  *   the #ifdef's are getting a bit cumbersome.
  *
@@ -58,9 +49,6 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
 #include 	<stdio.h>
 #include	<errno.h>
 #include	<setjmp.h>
-#ifndef s40
-#include	<time.h>
-#endif
 #include	<pwd.h>
 #if	defined(sv4) || defined(ptx)
 #include	<stropts.h>
@@ -72,7 +60,7 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
 #include	<sys/ioctl.h>
 #endif
 #else
-#if	defined(sv3) || defined(sgi) || defined(isc) || defined(ct)
+#if	defined(sv3) || defined(sgi) || defined(isc) || defined(ct) || defined(asv)
 #include	<termio.h>
 #if	defined(isc)
 #include	<sys/sioctl.h>
@@ -80,7 +68,7 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
 #endif
 #else
 #include	<sgtty.h>
-#endif	/* sv3 || sgi || isc */
+#endif
 #endif	/* POSIX */
 
 #include	"osdep.h"
@@ -97,6 +85,7 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
 #include	<sys/select.h>
 #endif
 
+int timeout = 0;
 
 /*
  * Immediately below are includes and declarations for the 3 basic
@@ -107,7 +96,7 @@ static char rcsid[] = "$Id: os_unix.c,v 4.51 1994/10/10 23:28:56 mikes Exp $";
 struct termios nstate,
 		ostate;
 #else
-#if	defined(sv3) || defined(sgi) || defined(isc)
+#if	defined(sv3) || defined(sgi) || defined(isc) || defined(asv)
 
 struct termio nstate,
               ostate;
@@ -120,7 +109,7 @@ struct  ltchars	nltchars = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
 struct  tchars	otchars;			/* old term special chars */
 struct  tchars	ntchars = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
 
-#endif	/* sv3 || sgi || isc */
+#endif	/* sv3 || sgi || isc || asv */
 #endif	/* POSIX */
 
 #if	defined(sv3) || defined(ct)
@@ -137,24 +126,37 @@ struct winsize {
 };
 #endif
 
-#ifdef	bsd
+#if	defined(bsd) || defined(lnx)
 int	errno;					/* ya, I know... */
 #endif
 
-#if	(defined(bsd) || defined(dyn) || defined(ct)) && !defined(LINUX)
+#if	defined(bsd) || defined(dyn) || defined(ct)
 #define	SIGTYPE int
 #else
 #define	SIGTYPE	void
 #endif
 
+#ifdef	SIGCHLD
+static jmp_buf pico_child_state;
+static short   pico_child_jmp_ok, pico_child_done;
+#endif
+
+#ifdef	MOUSE
+static int mexist = 0;		/* is the mouse driver installed? */
+static unsigned mnoop;
+#endif
 
 #ifdef	ANSI
-    int      kbseq(int *);
+    int      kbseq(KBESC_T *, int (*)(), int *);
     SIGTYPE  do_hup_signal();
     SIGTYPE  rtfrmshell();
 #ifdef	TIOCGWINSZ
     SIGTYPE  winch_handler();
 #endif
+#ifdef	SIGCHLD
+    SIGTYPE  child_handler();
+#endif
+
 #else
     int      kbseq();
     SIGTYPE  do_hup_signal();
@@ -162,8 +164,12 @@ int	errno;					/* ya, I know... */
 #ifdef	TIOCGWINSZ
     SIGTYPE  winch_handler();
 #endif
+#ifdef	SIGCHLD
+    SIGTYPE  child_handler();
 #endif
 
+#endif	/* ANSI */
+
 
 /*
  * for alt_editor arg[] building
@@ -188,7 +194,7 @@ ttopen()
 	nstate.c_cc[VTIME] = 0;
 	tcsetattr (0, TCSADRAIN, &nstate);
 #else
-#if	defined(sv3) || defined(sgi) || defined(isc) || defined(ct)
+#if	defined(sv3) || defined(sgi) || defined(isc) || defined(ct) || defined(asv)
 	(void) ioctl(0, TCGETA, &ostate);
 	(void) ioctl(0, TCGETA, &nstate);	/** again! **/
 
@@ -212,8 +218,44 @@ ttopen()
 	ioctl(0, TIOCSETC, &ntchars);		/* put new character into K */
 #endif	/* sv3 */
 #endif	/* POSIX */
+#ifdef	MOUSE
+	if(gmode & MDMOUSE)
+	  init_mouse();
+#endif	/* MOUSE */
     }
 
+#ifdef	MOUSE
+    if(mexist)
+      kpinsert(&pico_kbesc, "\033[M", K_XTERM_MOUSE);
+#endif	/* MOUSE */
+
+    if(!Pmaster || (gmode ^ MDEXTFB))
+      picosigs();
+
+    return(1);
+}
+
+
+/*
+ * ttresize - recompute the screen dimensions if necessary, and then
+ *	      adjust pico's internal buffers accordingly.
+ */
+int
+ttresize ()
+{
+    int row = -1, col = -1;
+
+    ttgetwinsz(&row, &col);
+    resize_pico(row, col);
+}
+
+
+/*
+ * picosigs - Install any handlers for the signals we're interested
+ *	      in catching.
+ */
+picosigs()
+{
     signal(SIGHUP,  do_hup_signal);	/* deal with SIGHUP */
     signal(SIGTERM, do_hup_signal);	/* deal with SIGTERM */
 #ifdef	SIGTSTP
@@ -222,11 +264,9 @@ ttopen()
 #ifdef	TIOCGWINSZ
     signal(SIGWINCH, winch_handler); /* window size changes */
 #endif
-    return(1);
 }
 
 
-
 /*
  * ttclose - this function gets called just before we go back home to 
  *           the command interpreter.  If called as pine composer, don't
@@ -248,7 +288,7 @@ ttclose()
 #ifdef	POSIX
 	tcsetattr (0, TCSADRAIN, &ostate);
 #else
-#if	defined(sv3) || defined(sgi) || defined(isc) || defined(ct)
+#if	defined(sv3) || defined(sgi) || defined(isc) || defined(ct) || defined(asv)
         ioctl(0, TCSETA, &ostate);
 #else
 	ioctl(0, TIOCSETP, &ostate);
@@ -267,8 +307,11 @@ ttclose()
 	 */
 	if(ostate.sg_ospeed <= B2400)
 	  sleep(1);
-#endif	/* sv3 || sgi || isc */
+#endif
 #endif	/* POSIX */
+#ifdef	MOUSE
+	end_mouse();
+#endif
     }
 
     return(1);
@@ -285,10 +328,14 @@ ttisslow()
 
     return((tcgetattr (1, &tty) == 0) ? cfgetospeed (&tty) < B9600 : FALSE);
 #else
-#if	defined(sv3) || defined(sgi) || defined(isc)
+#if	defined(sv3) || defined(sgi) || defined(isc) || defined(asv)
     struct termio tty;
 
+#ifdef asv
+    return((ioctl(1, TCGETA, &tty) == 0) ? (tty.c_cflag && CBAUD) < B9600 : FALSE);
+#else /* asv */
     return((tcgetattr (1, &tty) == 0) ? cfgetospeed (&tty) < B9600 : FALSE);
+#endif /* asv */
 #else
     struct  sgttyb tty;
 
@@ -299,16 +346,29 @@ ttisslow()
 
 
 /*
- * ttgetwinsz - set global rows and columns values and return
+ * ttgetwinsz - set global row and column values (if we can get them)
+ *		and return.
  */
-ttgetwinsz()
+ttgetwinsz(row, col)
+    int *row, *col;
 {
+    if(*row < 0)
+      *row = NROW - 1;
+    if(*col <= 0)
+      *col = NCOL;
 #ifdef TIOCGWINSZ
-    struct winsize win;
+    {
+	struct winsize win;
 
-    if (ioctl(0, TIOCGWINSZ, &win) == 0) {
-        term.t_ncol = (win.ws_col) ? win.ws_col : 80;
-        term.t_nrow = (win.ws_row) ? win.ws_row - 1 : 23;
+	if (ioctl(0, TIOCGWINSZ, &win) == 0) {	/* set to anything useful.. */
+	    if(win.ws_row)			/* ... the tty drivers says */
+	      *row = win.ws_row - 1;
+
+	    if(win.ws_col)
+	      *col = win.ws_col;
+	}
+
+	signal(SIGWINCH, winch_handler);	/* window size changes */
     }
 #endif
 }
@@ -343,14 +403,12 @@ ttgetc()
     unsigned char c;
     int i;
 
-    if((i = read(0, &c, 1)) <= 0){
-	if(i == 0 && errno == EINTR)		/* only acceptable failure */
-	  return(NODATA);
-	else
-	  kill(getpid(), SIGHUP);		/* fake a hup */
-    }
-    else
-      return((int)c);
+    if((i = read(0, &c, 1)) == 1)			/* success */
+      return((int) c);
+    else if(i == 0 || errno != EINTR)
+      kill(getpid(), SIGHUP);				/* eof or bad error */
+
+    return(NODATA);					/* acceptable error */
 }
 
 
@@ -376,7 +434,7 @@ typahead()
  */
 GetKey()
 {
-    int    c;
+    int    c, status, cc;
 
     if(timeout){
 	/*
@@ -416,57 +474,123 @@ GetKey()
 	  return(NODATA);		/* we really did time out */
     }
 
-    if ((c = (*term.t_getchar)()) == METACH) { /* Apply M- prefix      */
-	int status;
-	    
-	/*
-	 * this code should intercept special keypad keys
-	 */
-	switch(status = kbseq(&c)){
-	  case 0 : 	/* no dice */
-	    return(c);
-	  case  K_PAD_UP		:
-	  case  K_PAD_DOWN		:
-	  case  K_PAD_RIGHT		:
-	  case  K_PAD_LEFT		:
-	  case  K_PAD_PREVPAGE	:
-	  case  K_PAD_NEXTPAGE	:
-	  case  K_PAD_HOME		:
-	    return(status);
-	  case F1  :
-	  case F2  :
-	  case F3  :
-	  case F4  :
-	  case F5  :
-	  case F6  :
-	  case F7  :
-	  case F8  :
-	  case F9  :
-	  case F10 :
-	  case F11 :
-	  case F12 :
-	    return(status);
-	  case BADESC :
-	    if(c == '\033'){
-		c = (*term.t_getchar)();
-		if(islower(c))	/* canonicalize c */
-		  c = toupper(c);
+    switch(status = kbseq(pico_kbesc, term.t_getchar, &c)){
+      case 0: 	/* regular character */
+	break;
 
-		return((isalpha(c) || c == '@' || (c >= '[' && c <= '_'))
-		       ? (CTRL | c) : c);
+      case K_DOUBLE_ESC:
+	c = (*term.t_getchar)();
+
+	if(isdigit(c)){
+	    int n = 0, i = c - '0';
+
+	    if(!strchr("012", c))
+	      return(c);		/* bogus literal char value */
+
+	    while(n++ < 2){
+		c = (*term.t_getchar)();
+					/* remember Horner? */
+		if(!isdigit(c) || (n == 1 && i == 2 && c > '5'))
+		  return(c);		/* bogus literal char value */
+
+		i = (i * 10) + (c - '0');
 	    }
 
-	  default :				/* punt the whole thing	*/
-	    (*term.t_beep)();
+	    c = i;
 	    break;
 	}
+	else{
+	    if(islower(c))	/* canonicalize c */
+	      c = toupper(c);
+
+	    return((isalpha(c) || c == '@' || (c >= '[' && c <= '_'))
+		   ? (CTRL | c) : c);
+	}
+
+#ifdef MOUSE
+      case K_XTERM_MOUSE:
+	{
+	    static int down = 0;
+	    int        x, y, button;
+	    unsigned   ch;
+
+	    button = (*term.t_getchar)() & 0x03;
+	    c = (*term.t_getchar)();
+	    x = c - '!';
+	    c = (*term.t_getchar)();
+	    y = c - '!';
+	    if(button == 0){
+		down = 1;
+		if(checkmouse(&ch, 1, x, y))
+		  return(ch);
+	    }
+	    else if(down && button == 3){
+		down = 0;
+		if(checkmouse(&ch, 0, x, y))
+		  return(ch);
+	    }
+
+	    return(NODATA);
+	}
+
+	break;
+#endif /* MOUSE */
+
+      case  K_PAD_UP		:
+      case  K_PAD_DOWN		:
+      case  K_PAD_RIGHT		:
+      case  K_PAD_LEFT		:
+      case  K_PAD_PREVPAGE	:
+      case  K_PAD_NEXTPAGE	:
+      case  K_PAD_HOME		:
+      case  K_PAD_END		:
+      case  K_PAD_DELETE	:
+      case F1  :
+      case F2  :
+      case F3  :
+      case F4  :
+      case F5  :
+      case F6  :
+      case F7  :
+      case F8  :
+      case F9  :
+      case F10 :
+      case F11 :
+      case F12 :
+	return(status);
+
+      case K_SWALLOW_TIL_Z:
+	status = NODATA;
+      case K_SWALLOW_UP:
+      case K_SWALLOW_DOWN:
+      case K_SWALLOW_LEFT:
+      case K_SWALLOW_RIGHT:
+	for(cc = 0; cc != 'z' && cc != '~';)
+	  cc = (*term.t_getchar)() & 0x7f;
+
+	return(status);
+	break;
+
+      case K_KERMIT:
+	for(cc = 0; cc != '\033' || ((*term.t_getchar)() & 0x7f) != '\\';)
+	  cc = (*term.t_getchar)() & 0x7f;
+
+	c = NODATA;
+	break;
+
+      case BADESC:
+	(*term.t_beep)();
+	return(status);
+
+      default:				/* punt the whole thing	*/
+	(*term.t_beep)();
+	break;
     }
 
     if (c>=0x00 && c<=0x1F)                 /* C0 control -> C-     */
       c = CTRL | (c+'@');
 
     return (c);
-
 }
 
 
@@ -476,27 +600,29 @@ GetKey()
  *         compares it to a trie of known keyboard escape sequences, and
  *         performs the function bound to the escape sequence.
  * 
- *         returns: BADESC, the escaped function, or 0 if not found.
+ *         returns: BADESC, the escaped function, or 0 if a regular char.
  */
-kbseq(c)
-int	*c;
+kbseq(trie, getcfunc, c)
+    KBESC_T *trie;
+    int    (*getcfunc)();
+    int	    *c;
 {
-    register char	b;
-    register int	first = 1;
-    register struct	KBSTREE	*current = kpadseqs;
+    register char     b;
+    register int      first = 1;
+    register KBESC_T *current = trie;
 
-    if(kpadseqs == NULL)			/* bag it */
+    if(trie == NULL)				/* bag it */
       return(BADESC);
 
     while(1){
-	*c = b = (*term.t_getchar)();
+	b = *c = (*getcfunc)();
 
 	while(current->value != b){
 	    if(current->left == NULL){		/* NO MATCH */
 		if(first)
-		  return(BADESC);
+		  return(0);			/* regular char */
 		else
-		  return(0);
+		  return(BADESC);
 	    }
 	    current = current->left;
 	}
@@ -512,6 +638,111 @@ int	*c;
 
 
 
+#define	newnode()	(KBESC_T *)malloc(sizeof(KBESC_T))
+/*
+ * kpinsert - insert a keystroke escape sequence into the global search
+ *	      structure.
+ */
+void
+kpinsert(trie, kstr, kval)
+    KBESC_T **trie;
+    char     *kstr;
+    int       kval;
+{
+    register	char	*buf;
+    register	KBESC_T *temp;
+    register	KBESC_T *trail;
+
+    if(kstr == NULL)
+      return;
+
+#ifndef TERMCAP_WINS
+    /*
+     * Don't allow escape sequences that don't start with ESC unless
+     * TERMCAP_WINS is defined.  This is to protect against mistakes
+     * in termcap files.
+     */
+    if(*kstr != '\033')
+      return;
+#endif /* !TERMCAP_WINS */
+
+    temp = trail = *trie;
+    buf = kstr;
+
+    for(;;){
+	if(temp == NULL){
+	    temp = newnode();
+	    temp->value = *buf;
+	    temp->func = 0;
+	    temp->left = NULL;
+	    temp->down = NULL;
+	    if(*trie == NULL)
+	      *trie = temp;
+	    else
+	      trail->down = temp;
+	}
+	else{				/* first entry */
+	    while((temp != NULL) && (temp->value != *buf)){
+		trail = temp;
+		temp = temp->left;
+	    }
+
+	    if(temp == NULL){   /* add new val */
+		temp = newnode();
+		temp->value = *buf;
+		temp->func = 0;
+		temp->left = NULL;
+		temp->down = NULL;
+		trail->left = temp;
+	    }
+	}
+
+	if(*(++buf) == '\0')
+	  break;
+	else{
+	    /*
+	     * Ignore attempt to overwrite shorter existing escape sequence.
+	     * That means that sequences with higher priority should be
+	     * set up first, so if we want termcap sequences to override
+	     * hardwired sequences, put the kpinsert calls for the
+	     * termcap sequences first.  (That's what you get if you define
+	     * TERMCAP_WINS.)
+	     */
+	    if(temp->func != 0)
+	      return;
+
+	    trail = temp;
+	    temp = temp->down;
+	}
+    }
+    
+    /*
+     * Ignore attempt to overwrite longer sequences we are a prefix
+     * of (down != NULL) and exact same sequence (func != 0).
+     */
+    if(temp != NULL && temp->down == NULL && temp->func == 0)
+      temp->func = kval;
+}
+
+
+
+/*
+ * kbdestroy() - kills the key pad function key search tree
+ *		 and frees all lines associated with it
+ */
+void
+kbdestroy(kb)
+    KBESC_T *kb;
+{
+    if(kb){
+	kbdestroy(kb->left);
+	kbdestroy(kb->down);
+	free((char *) kb);
+    }
+}
+
+
+
 /*
  * alt_editor - fork off an alternate editor for mail message composition 
  *              if one is configured and passed from pine.  If not, only
@@ -524,68 +755,91 @@ alt_editor(f, n)
     char   *fn;					/* tmp holder for file name */
     char   *cp;
     char   *args[MAXARGS];			/* ptrs into edit command */
-    char   *writetmp();
+    char   result[128];				/* result string */
     int	   child, pid, i, done = 0;
     long   l;
-#if	defined(POSIX) || defined(sv3) || defined(COHERENT) || defined(isc) || defined(neb)
-    int    stat;
-#else
-    union  wait stat;
-#endif
     FILE   *p;
     SIGTYPE (*ohup)(), (*oint)(), (*osize)(), (*ostop)(), (*ostart)();
+#if	defined(POSIX) || defined(sv3) || defined(COHERENT) || defined(isc) || defined(neb)
+    int    stat;
+#ifndef	WIFEXITED
+#define	WIFEXITED(X)	(!((X) & 0xff))		/* low bits, child killed */
+#endif
+#ifndef	WEXITSTATUS
+#define	WEXITSTATUS(X)	((X) >> 8)		/* high bits, exit value */
+#endif
+#else
+    union  wait stat;
+#ifndef	WIFEXITED
+#define	WIFEXITED(X)	(!(X).w_termsig)	/* nonzero if child killed */
+#endif
+#ifndef	WEXITSTATUS
+#define	WEXITSTATUS(X)	X.w_retcode		/* child's exit value */
+#endif
+#endif
 
     if(Pmaster == NULL)
       return(-1);
 
     if(gmode&MDSCUR){
-	emlwrite("Alternate editor not available in restricted mode", NULL);
+	emlwrite("Alternate %s not available in restricted mode",
+		 f ? "speller" : "editor");
 	return(-1);
     }
 
-    if(Pmaster->alt_ed == NULL){
-	if(!(gmode&MDADVN)){
-	    emlwrite("\007Unknown Command",NULL);
-	    return(-1);
-	}
+    strcpy(result, "Alternate %s complete.");
 
-	if(getenv("EDITOR"))
-	  strcpy(eb, (char *)getenv("EDITOR"));
+    if(f){
+	if(alt_speller)
+	  strcpy(eb, alt_speller);
 	else
-	  *eb = '\0';
-
-	while(!done){
-	    pid = mlreplyd("Which alternate editor ? ",eb,NLINE,QDEFLT,NULL);
-
-	    switch(pid){
-	      case ABORT:
+	  return(-1);
+    }
+    else{
+	if(Pmaster->alt_ed == NULL){
+	    if(!(gmode&MDADVN)){
+		emlwrite("\007Unknown Command",NULL);
 		return(-1);
-	      case HELPCH:
-		emlwrite("no alternate editor help yet", NULL);
+	    }
+
+	    if(getenv("EDITOR"))
+	      strcpy(eb, (char *)getenv("EDITOR"));
+	    else
+	      *eb = '\0';
+
+	    while(!done){
+		pid = mlreplyd("Which alternate editor ? ", eb,
+			       NLINE, QDEFLT, NULL);
+		switch(pid){
+		  case ABORT:
+		    return(-1);
+		  case HELPCH:
+		    emlwrite("no alternate editor help yet", NULL);
 
 /* take sleep and break out after there's help */
-		sleep(3);
-		break;
-	      case (CTRL|'L'):
-		sgarbf = TRUE;
-		update();
-		break;
-	      case TRUE:
-	      case FALSE:			/* does editor exist ? */
-		if(*eb == '\0'){		/* leave silently? */
-		    mlerase();
-		    return(-1);
-		}
+		    sleep(3);
+		    break;
+		  case (CTRL|'L'):
+		    sgarbf = TRUE;
+		    update();
+		    break;
+		  case TRUE:
+		  case FALSE:			/* does editor exist ? */
+		    if(*eb == '\0'){		/* leave silently? */
+			mlerase();
+			return(-1);
+		    }
 
-		done++;
-		break;
-	      default:
-		break;
+		    done++;
+		    break;
+		    default:
+		    break;
+		}
 	    }
 	}
+	else
+	  strcpy(eb, Pmaster->alt_ed);
     }
-    else
-      strcpy(eb, Pmaster->alt_ed);
 
     if((fn=writetmp(0, 1)) == NULL){		/* get temp file */
 	emlwrite("Problem writing temp file for alt editor", NULL);
@@ -625,23 +879,54 @@ alt_editor(f, n)
 
     args[i] = NULL;
 
+    for(i = 0; i <= ((Pmaster) ? term.t_nrow : term.t_nrow - 1); i++){
+	movecursor(i, 0);
+	if(!i){
+	    fputs("Invoking alternate ", stdout);
+	    fputs(f ? "speller..." : "editor...", stdout);
+	}
+
+	peeol();
+    }
+
+    (*term.t_flush)();
     if(Pmaster)
       (*Pmaster->raw_io)(0);			/* turn OFF raw mode */
 
-    emlwrite("Invoking alternate editor...", NULL);
+#ifdef	SIGCHLD
+    if(Pmaster){
+	/*
+	 * The idea here is to keep any mail connections our caller
+	 * may have open while our child's out running around...
+	 */
+	pico_child_done = pico_child_jmp_ok = 0;
+	(void) signal(SIGCHLD,  child_handler);
+    }
+#endif
 
-    if(child=fork()){			/* wait for the child to finish */
+    if((child = fork()) > 0){		/* wait for the child to finish */
 	ohup = signal(SIGHUP, SIG_IGN);	/* ignore signals for now */
 	oint = signal(SIGINT, SIG_IGN);
 #ifdef	TIOCGWINSZ
         osize = signal(SIGWINCH, SIG_IGN);
 #endif
 
-/*
- * BUG - wait should be made non-blocking and mail_pings or something 
- * need to be done in the loop to keep the imap stream alive
- */
-	while((pid=(int)wait(&stat)) != child)
+#ifdef	SIGCHLD
+	if(Pmaster){
+	    while(!pico_child_done){
+		(*Pmaster->newmail)(0, 0);
+		if(!pico_child_done)
+		  if(setjmp(pico_child_state) == 0){
+		      pico_child_jmp_ok = 1;
+		      sleep(600);
+		  }
+
+		pico_child_jmp_ok = 0;
+	    }
+	}
+#endif
+
+	while((pid = (int) wait(&stat)) != child)
 	  ;
 
 	signal(SIGHUP, ohup);	/* restore signals */
@@ -649,16 +934,34 @@ alt_editor(f, n)
 #ifdef	TIOCGWINSZ
         signal(SIGWINCH, osize);
 #endif
+
+	/*
+	 * Report child's unnatural or unhappy exit...
+	 */
+	if(WIFEXITED(stat) && WEXITSTATUS(stat) == 0)
+	  strcpy(result, "Alternate %s done");
+	else
+	  sprintf(result, "Alternate %%s abnormally terminated (%d)",
+		  WIFEXITED(stat) ? WEXITSTATUS(stat) : -1);
     }
-    else{				/* spawn editor */
+    else if(child == 0){		/* spawn editor */
 	signal(SIGHUP, SIG_DFL);	/* let editor handle signals */
 	signal(SIGINT, SIG_DFL);
 #ifdef	TIOCGWINSZ
         signal(SIGWINCH, SIG_DFL);
+#endif
+#ifdef	SIGCHLD
+	(void) signal(SIGCHLD,  SIG_DFL);
 #endif
 	if(execvp(args[0], args) < 0)
-	  exit(1);
+	  exit(-1);
     }
+    else				/* error! */
+      sprintf(result, "\007Can't fork %%s: %s", errstr(errno));
+
+#ifdef	SIGCHLD
+    (void) signal(SIGCHLD,  SIG_DFL);
+#endif
 
     if(Pmaster)
       (*Pmaster->raw_io)(1);		/* turn ON raw mode */
@@ -672,6 +975,8 @@ alt_editor(f, n)
     curbp->b_flag |= BFCHG;		/* mark dirty for packbuf() */
     ttopen();				/* reset the signals */
     refresh(0, 1);			/* redraw */
+    update();
+    emlwrite(result, f ? "speller" : "editor");
     return(0);
 }
 
@@ -690,23 +995,51 @@ bktoshell()		/* suspend MicroEMACS and wait to wake up */
 	return;
     }
 
-    if(Pmaster){
-	(*Pmaster->raw_io)(0);	/* actually in pine source */
+    if(gmode&MDSPWN){
+	char *shell;
+
+	if(Pmaster){
+	    (*Pmaster->raw_io)(0);	/* actually in pine source */
+
+	    movecursor(0, 0);
+	    (*term.t_eeop)();
+	    printf("\n\n\nUse \"exit\" to return to Pine\n");
+	}
+	else{
+	    vttidy();
+	    movecursor(0, 0);
+	    (*term.t_eeop)();
+	    printf("\n\n\nUse \"exit\" to return to Pico\n");
+	}
+
+	system((shell = (char *)getenv("SHELL")) ? shell : "/bin/csh");
+	rtfrmshell();			/* fixup tty */
+    }
+    else {
+	if(Pmaster){
+	    (*Pmaster->raw_io)(0);	/* actually in pine source */
+
+	    movecursor(term.t_nrow, 0);
+	    printf("\n\n\nUse \"fg\" to return to Pine\n");
+	}
+	else{
+	    movecursor(term.t_nrow-1, 0);
+	    peeol();
+	    movecursor(term.t_nrow, 0);
+	    peeol();
+	    movecursor(term.t_nrow, 0);
+	    printf("\n\n\nUse \"fg\" to return to Pico\n");
+	    ttclose();
+	}
 
 	movecursor(term.t_nrow, 0);
-	printf("\n\n\nUse \"fg\" to return to Pine\n");
+	peeol();
+	(*term.t_flush)();
 
+	signal(SIGCONT, rtfrmshell);	/* prepare to restart */
+	signal(SIGTSTP, SIG_DFL);			/* prepare to stop */
+	kill(0, SIGTSTP);
     }
-    else
-      vttidy();
-
-    movecursor(term.t_nrow, 0);
-    peeol();
-    (*term.t_flush)();
-
-    signal(SIGCONT, rtfrmshell);	/* prepare to restart */
-    signal(SIGTSTP, SIG_DFL);			/* prepare to stop */
-    kill(0, SIGTSTP);
 #endif
 }
 
@@ -726,24 +1059,7 @@ rtfrmshell()
     }
 
     ttopen();
-
-#ifdef	TIOCGWINSZ
-    {
-	struct winsize win;
-	extern int resize_pico();
-
-	/*
-	 * refit pico of window size changed....
-	 */
-	if (ioctl(0, TIOCGWINSZ, &win) == 0) {
-	    if (win.ws_col && win.ws_row)
-	      resize_pico(win.ws_row - 1, win.ws_col);
-	}
-    }
-#endif
-
-    sgarbf = TRUE;
-    curwp->w_flag = WFHARD;
+    ttresize();
     refresh(0, 1);
 #endif
 }
@@ -790,7 +1106,7 @@ do_hup_signal()
 unsigned char okinfname[32] = {
       0,    0, 			/* ^@ - ^G, ^H - ^O  */
       0,    0,			/* ^P - ^W, ^X - ^_  */
-      0,    0x17,		/* SP - ' ,  ( - /   */
+      0x80, 0x17,		/* SP - ' ,  ( - /   */
       0xff, 0xc0,		/*  0 - 7 ,  8 - ?   */
       0x7f, 0xff,		/*  @ - G ,  H - O   */
       0xff, 0xe1,		/*  P - W ,  X - _   */
@@ -820,15 +1136,22 @@ char c;
  */
 fexist(file, m, l)
 char *file;
-char *m;					/* files mode: r, w or rw */
-long *l;
+char *m;			/* files mode: r,w,rw,t or x */
+long *l;			/* t means use lstat         */
 {
     struct stat	sbuf;
+    extern int lstat();
+    int		(*stat_f)() = (m && *m == 't') ? lstat : stat;
+    int accessible;
+/* define them here so we don't have to worry about what to include */
+#define EXECUTE_ACCESS  01
+#define WRITE_ACCESS    02
+#define READ_ACCESS     04
 
     if(l)
       *l = 0L;
-
-    if(stat(file, &sbuf) < 0){
+    
+    if((*stat_f)(file, &sbuf) < 0){
 	switch(errno){
 	  case ENOENT :				/* File not found */
 	    return(FIOFNF);
@@ -836,6 +1159,8 @@ long *l;
 	  case ENAMETOOLONG :			/* Name is too long */
 	    return(FIOLNG);
 #endif
+	  case EACCES :				/* File not found */
+	    return(FIOPER);
 	  default:				/* Some other error */
 	    return(FIOERR);
 	}
@@ -846,14 +1171,51 @@ long *l;
 
     if((sbuf.st_mode&S_IFMT) == S_IFDIR)
       return(FIODIR);
+    else if(*m == 't'){
+	struct stat	sbuf2;
 
-    if(m[0] == 'r')				/* read access? */
-      return((S_IREAD&sbuf.st_mode) ? FIOSUC : FIONRD);
-    else if(m[0] == 'w')			/* write access? */
-      return((S_IWRITE&sbuf.st_mode) ? FIOSUC : FIONWT);
-    else if(m[0] == 'x')			/* execute access? */
-      return((S_IEXEC&sbuf.st_mode) ? FIOSUC : FIONEX);
-    return(FIOERR);				/* what? */
+	/*
+	 * If it is a symbolic link pointing to a directory, treat
+	 * it like it is a directory, not a link.
+	 */
+	if((sbuf.st_mode&S_IFMT) == S_IFLNK){
+	    if(stat(file, &sbuf2) < 0){
+		switch(errno){
+		  case ENOENT :				/* File not found */
+		    return(FIOSYM);			/* call it a link */
+#ifdef	ENAMETOOLONG
+		  case ENAMETOOLONG :			/* Name is too long */
+		    return(FIOLNG);
+#endif
+		  case EACCES :				/* File not found */
+		    return(FIOPER);
+		  default:				/* Some other error */
+		    return(FIOERR);
+		}
+	    }
+
+	    if((sbuf2.st_mode&S_IFMT) == S_IFDIR)
+	      return(FIODIR);
+	}
+
+	return(((sbuf.st_mode&S_IFMT) == S_IFLNK) ? FIOSYM : FIOSUC);
+    }
+
+    if(*m == 'r'){				/* read access? */
+	if(*(m+1) == 'w')			/* and write access? */
+	  return((access(file,READ_ACCESS)==0)
+		 ? (access(file,WRITE_ACCESS)==0)
+		    ? FIOSUC
+		    : FIONWT
+		 : FIONRD);
+	else if(!*(m+1))			/* just read access? */
+	  return((access(file,READ_ACCESS)==0) ? FIOSUC : FIONRD);
+    }
+    else if(*m == 'w' && !*(m+1))		/* write access? */
+      return((access(file,WRITE_ACCESS)==0) ? FIOSUC : FIONWT);
+    else if(*m == 'x' && !*(m+1))		/* execute access? */
+      return((access(file,EXECUTE_ACCESS)==0) ? FIOSUC : FIONEX);
+    return(FIOERR);				/* bad m arg */
 }
 
 
@@ -966,8 +1328,8 @@ int err;
  *             malloc'd string.  n contains the number of names
  */
 char *
-getfnames(dn, n)
-char *dn;
+getfnames(dn, pat, n, e)
+char *dn, *pat, *e;
 int  *n;
 {
     long           l;
@@ -979,7 +1341,7 @@ int  *n;
 #else
     DIR           *dirp;			/* opened directory */
 #endif
-#if	defined(POSIX) || defined(aix) || defined(COHERENT) || defined(isc) || defined(sv3)
+#if	defined(POSIX) || defined(aix) || defined(COHERENT) || defined(isc) || defined(sv3) || defined(asv)
     struct dirent *dp;
 #else
     struct direct *dp;
@@ -990,15 +1352,21 @@ int  *n;
     if(stat(dn, &sbuf) < 0){
 	switch(errno){
 	  case ENOENT :				/* File not found */
-	    emlwrite("\007File not found: \"%s\"", dn);
+	    if(e)
+	      sprintf(e, "\007File not found: \"%s\"", dn);
+
 	    break;
 #ifdef	ENAMETOOLONG
 	  case ENAMETOOLONG :			/* Name is too long */
-	    emlwrite("\007File name too long: \"%s\"", dn);
+	    if(e)
+	      sprintf(e, "\007File name too long: \"%s\"", dn);
+
 	    break;
 #endif
 	  default:				/* Some other error */
-	    emlwrite("\007Error getting file info: \"%s\"", dn);
+	    if(e)
+	      sprintf(e, "\007Error getting file info: \"%s\"", dn);
+
 	    break;
 	}
 	return(NULL);
@@ -1006,21 +1374,25 @@ int  *n;
     else{
 	l = sbuf.st_size;
 	if((sbuf.st_mode&S_IFMT) != S_IFDIR){
-	    emlwrite("\007Not a directory: \"%s\"", dn);
+	    if(e)
+	      sprintf(e, "\007Not a directory: \"%s\"", dn);
+
 	    return(NULL);
 	}
     }
 
     if((names=(char *)malloc(sizeof(char)*l)) == NULL){
-	emlwrite("\007Can't malloc space for file names", NULL);
+	if(e)
+	  sprintf(e, "\007Can't malloc space for file names", NULL);
+
 	return(NULL);
     }
 
     errno = 0;
     if((dirp=opendir(dn)) == NULL){
-	char buf[NLINE];
-	sprintf(buf,"\007Can't open \"%s\": %s", dn, errstr(errno));
-	emlwrite(buf, NULL);
+	if(e)
+	  sprintf(e, "\007Can't open \"%s\": %s", dn, errstr(errno));
+
 	free((char *)names);
 	return(NULL);
     }
@@ -1040,12 +1412,13 @@ int  *n;
 	  ;
     }
 #else
-    while((dp = readdir(dirp)) != NULL){
-	(*n)++;
-	p = dp->d_name;
-	while((*np++ = *p++) != '\0')
-	  ;
-    }
+    while((dp = readdir(dirp)) != NULL)
+      if(!pat || !*pat || !strncmp(dp->d_name, pat, strlen(pat))){
+	  (*n)++;
+	  p = dp->d_name;
+	  while(*np++ = *p++)
+	    ;
+      }
 #endif
 
     closedir(dirp);					/* shut down */
@@ -1080,6 +1453,9 @@ char *f;
       case FIONRD:
 	emlwrite("\007Read permission denied: %s", f);
 	break;
+      case FIOPER:
+	emlwrite("\007Permission denied: %s", f);
+	break;
       case FIONEX:
 	emlwrite("\007Execute permission denied: %s", f);
 	break;
@@ -1100,26 +1476,36 @@ int  len;
 {
     struct passwd *pw;
     register char *x, *y, *z;
+    char *home = NULL;
     char name[20];
     
     if(*fn == '~') {
-        for(x = fn+1, y = name; *x != '/' && *x != '\0'; *y++ = *x++);
+        for(x = fn+1, y = name; *x != '/' && *x != '\0'; *y++ = *x++)
+	  ;
+
         *y = '\0';
-        if(x == fn + 1) 
-          pw = getpwuid(getuid());
-        else
-          pw = getpwnam(name);
-        if(pw == NULL)
-          return(NULL);
-        if(strlen(pw->pw_dir) + strlen(fn) > len) {
-            return(NULL);
-        }
+        if(x == fn + 1){			/* ~/ */
+	    if (!(home = (char *) getenv("HOME")))
+	      if (pw = getpwuid(geteuid()))
+		home = pw->pw_dir;
+	}
+	else if(*name){				/* ~username/ */
+	    if(pw = getpwnam(name))
+	      home = pw->pw_dir;
+	}
+
+        if(!home || (strlen(home) + strlen(fn) > len))
+	  return(NULL);
+
 	/* make room for expanded path */
-	for(z=x+strlen(x),y=fn+strlen(x)+strlen(pw->pw_dir);
+	for(z = x + strlen(x), y = fn + strlen(x) + strlen(home);
 	    z >= x;
-	    *y-- = *z--);
+	    *y-- = *z--)
+	  ;
+
 	/* and insert the expanded address */
-	for(x=fn,y=pw->pw_dir; *y != '\0'; *x++ = *y++);
+	for(x = fn, y = home; *y != '\0'; *x++ = *y++)
+	  ;
     }
     return(fn);
 }
@@ -1142,11 +1528,23 @@ int  len;
 	if(Pmaster && !(gmode&MDCURDIR)
                    && (*name != '~' && strlen(name)+2 <= len)){
 
-	    for(shft = strchr(name, '\0'); shft >= name; shft--)
-	      shft[2] = *shft;
+	    if(gmode&MDTREE && strlen(name)+strlen(opertree)+1 <= len){
+		int i, off = strlen(opertree);
 
-	    name[0] = '~';
-	    name[1] = '/';
+		for(shft = strchr(name, '\0'); shft >= name; shft--)
+		  shft[off] = *shft;
+		
+		off--;
+		for(shft=name, i=0; i < off; i++)
+		  shft[i] = opertree[i];
+	    }
+	    else{
+		for(shft = strchr(name, '\0'); shft >= name; shft--)
+		  shft[2] = *shft;
+
+		name[0] = '~';
+		name[1] = '/';
+	    }
 	}
 
 	pfnexpand(name, len);
@@ -1484,8 +1882,8 @@ time_to_check()
     if(!timeout)
       return(FALSE);
 
-    if(time((long *) 0) - lasttime > (time_t)timeout){
-	lasttime = time((long *) 0);
+    if(time((time_t *) 0) - lasttime > (time_t)timeout){
+	lasttime = time((time_t *) 0);
 	return(TRUE);
     }
     else
@@ -1511,25 +1909,73 @@ QcompType *s1, *s2;
 }
 
 
+/*
+ * chkptinit -- initialize anything we need to support composer
+ *		checkpointing
+ */
+chkptinit(file, n)
+    char *file;
+    int   n;
+{
+    unsigned pid;
+    char    *chp;
+
+    if(!file[0]){
+	long gmode_save = gmode;
+
+	if(gmode&MDCURDIR)
+	  gmode &= ~MDCURDIR;  /* so fixpath will use home dir */
+
+	strcpy(file, "#picoXXXXX#");
+	fixpath(file, NLINE);
+	gmode = gmode_save;
+    }
+    else{
+	int l = strlen(file);
+
+	if(file[l-1] != '/'){
+	    file[l++] = '/';
+	    file[l]   = '\0';
+	}
+
+	strcpy(file + l, "#picoXXXXX#");
+    }
+
+    pid = (unsigned)getpid();
+    for(chp = file+strlen(file) - 2; *chp == 'X'; chp--){
+	*chp = (pid % 10) + '0';
+	pid /= 10;
+    }
+
+    unlink(file);
+}
+
+
 #ifdef	TIOCGWINSZ
 /*
  * winch_handler - handle window change signal
  */
 SIGTYPE winch_handler()
 {
-    struct winsize win;
-    extern int resize_pico();
-
     signal(SIGWINCH, winch_handler);
-
-    if (ioctl(0, TIOCGWINSZ, &win) == 0) {
-	if (win.ws_col && win.ws_row)
-	  resize_pico(win.ws_row - 1, win.ws_col);
-    }
+    ttresize();
 }
 #endif	/* TIOCGWINSZ */
 
 
+#ifdef	SIGCHLD
+/*
+ * child_handler - handle child status change signal
+ */
+SIGTYPE child_handler()
+{
+    pico_child_done = 1;
+    if(pico_child_jmp_ok)
+      longjmp(pico_child_state, 1);
+}
+#endif	/* SIGCHLD */
+
+
 #if	defined(sv3) || defined(ct)
 /* Placed by rll to handle the rename function not found in AT&T */
 rename(oldname, newname)
@@ -1549,3 +1995,179 @@ rename(oldname, newname)
     return(rtn);
 }
 #endif
+
+
+#ifdef	MOUSE
+
+/* 
+ * init_mouse - check for xterm and initialize mouse tracking if present...
+ */
+init_mouse()
+{
+    if(mexist)
+      return(TRUE);
+
+    if(getenv("DISPLAY")){
+	mouseon();
+	return(mexist = TRUE);
+    }
+    else
+      return(FALSE);
+}
+
+
+/* 
+ * end_mouse - clear xterm mouse tracking if present...
+ */
+void
+end_mouse()
+{
+    if(mexist){
+	mexist = 0;			/* just see if it exists here. */
+	mouseoff();
+    }
+}
+
+
+/*
+ * mouseexist - function to let outsiders know if mouse is turned on
+ *              or not.
+ */
+mouseexist()
+{
+    return(mexist);
+}
+
+
+/*
+ * mouseon - call made available for programs calling pico to turn ON the
+ *           mouse cursor.
+ */
+void
+mouseon()
+{
+    fputs(XTERM_MOUSE_ON, stdout);
+}
+
+
+/*
+ * mouseon - call made available for programs calling pico to turn OFF the
+ *           mouse cursor.
+ */
+void
+mouseoff()
+{
+    fputs(XTERM_MOUSE_OFF, stdout);
+}
+
+
+/* 
+ * checkmouse - look for mouse events in key menu and return 
+ *              appropriate value.
+ */
+int
+checkmouse(ch, down, mcol, mrow)
+unsigned *ch;
+int	  down, mcol, mrow;
+{
+    static int oindex;
+    register int k;			/* current bit/button of mouse */
+    int i = 0, rv = 0;
+    MENUITEM *mp;
+
+    if(!mexist || mcol < 0 || mrow < 0)
+      return(FALSE);
+
+    if(down)			/* button down */
+      oindex = -1;
+
+    for(mp = mfunc; mp; mp = mp->next)
+      if(mp->action && M_ACTIVE(mrow, mcol, mp))
+	break;
+
+    if(mp){
+	unsigned long r;
+
+	r = (*mp->action)(down ? M_EVENT_DOWN : M_EVENT_UP,
+			  mrow, mcol, M_BUTTON_LEFT, 0);
+	if(r & 0xffff){
+	    *ch = (unsigned)((r>>16)&0xffff);
+	    rv  = TRUE;
+	}
+    }
+    else{
+	while(1){			/* see if we understand event */
+	    if(i >= 12){
+		i = -1;
+		break;
+	    }
+
+	    if(M_ACTIVE(mrow, mcol, &menuitems[i]))
+	      break;
+
+	    i++;
+	}
+
+	if(down){			/* button down */
+	    oindex = i;			/* remember where */
+	    if(i != -1
+	       && menuitems[i].label_hiliter != NULL
+	       && menuitems[i].val != mnoop)  /* invert label */
+	      (*menuitems[i].label_hiliter)(1, &menuitems[i]);
+	}
+	else{				/* button up */
+	    if(oindex != -1){
+		if(i == oindex){
+		    *ch = menuitems[i].val;
+		    rv = TRUE;
+		}
+	    }
+	}
+    }
+
+    /* restore label */
+    if(!down
+       && oindex != -1
+       && menuitems[oindex].label_hiliter != NULL
+       && menuitems[oindex].val != mnoop)
+      (*menuitems[oindex].label_hiliter)(0, &menuitems[oindex]);
+
+    return(rv);
+}
+
+
+/*
+ * invert_label - highlight the label of the given menu item.
+ */
+void
+invert_label(state, m)
+int state;
+MENUITEM *m;
+{
+    unsigned i, j;
+    int   r, c, p, oldr, oldc, col_offset;
+    char *lp;
+
+    /*
+     * Leave the command name bold
+     */
+    col_offset = (state || !(lp=strchr(m->label, ' '))) ? 0 : (lp - m->label);
+    movecursor((int)m->tl.r, (int)m->tl.c + col_offset);
+    (*term.t_rev)(state);
+
+    for(i = m->tl.r; i <= m->br.r; i++)
+      for(j = m->tl.c + col_offset; j <= m->br.c; j++)
+	if(i == m->lbl.r && j == m->lbl.c + col_offset && m->label){
+	    lp = m->label + col_offset;		/* show label?? */
+	    while(*lp && j++ < m->br.c)
+	      putc(*lp++, stdout);
+
+	    continue;
+	}
+	else
+	  putc(' ', stdout);
+
+    if(state)
+      (*term.t_rev)(0);  /* turn inverse back off */
+}
+#endif	/* MOUSE */
