@@ -1,5 +1,5 @@
 /*
- * $Id: os_unix.h,v 4.22 1994/10/10 23:29:12 mikes Exp $
+ * $Id: os_unix.h,v 4.32 1996/03/17 23:02:46 mikes Exp $
  *
  * Program:	Operating system dependent routines - Ultrix 4.1
  *
@@ -14,29 +14,16 @@
  *
  * Please address all bugs and comments to "pine-bugs@cac.washington.edu"
  *
- * Copyright 1991-1994  University of Washington
- *
- *  Permission to use, copy, modify, and distribute this software and its
- * documentation for any purpose and without fee to the University of
- * Washington is hereby granted, provided that the above copyright notice
- * appears in all copies and that both the above copyright notice and this
- * permission notice appear in supporting documentation, and that the name
- * of the University of Washington not be used in advertising or publicity
- * pertaining to distribution of the software without specific, written
- * prior permission.  This software is made available "as is", and
- * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
- * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
- * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
- * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
- * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
- * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- * Pine and Pico are trademarks of the University of Washington.
- * No commercial use of these trademarks may be made without prior
- * written permission of the University of Washington.
  *
+ * Pine and Pico are registered trademarks of the University of Washington.
+ * No commercial use of these trademarks may be made without prior written
+ * permission of the University of Washington.
+ * 
+ * Pine, Pico, and Pilot software and its included text are Copyright
+ * 1989-1996 by the University of Washington.
+ * 
+ * The full text of our legal notices is contained in the file called
+ * CPYRIGHT, included with this distribution.
  */
 
 #ifndef	OSDEP_H
@@ -65,12 +52,18 @@
 #include	<ctype.h>
 #include	<sys/types.h>
 
-#if	defined(POSIX) || defined(aix) || defined(COHERENT) || defined(isc) || defined(sv3)
+#if	defined(POSIX) || defined(aix) || defined(COHERENT) || defined(isc) || defined(sv3) || defined(asv)
 #include	<dirent.h>
 #else
 #include	<sys/dir.h>
 #endif
 
+#if	defined(asv)
+#include	<sys/socket.h>  /* for fd_set */
+extern struct passwd *getpwnam();
+extern struct passwd *getpwuid();
+#endif
+
 #if	defined(sco)
 #include	<sys/stream.h>
 #include	<sys/ptem.h>
@@ -78,6 +71,10 @@
 
 #include	<sys/stat.h>
 
+#if	defined(asv) || defined(bsd)
+extern int stat();
+#endif
+
 /*
  * 3b1 definition requirements
  */
@@ -94,7 +91,7 @@ typedef struct
 #endif
 
 /* Machine/OS definition			*/
-#if	defined(ptx) || defined(sgi) || defined(sv4)
+#if	defined(ptx) || defined(sgi) || defined(sv4) || defined(sco)
 #define TERMINFO	1               /* Use TERMINFO                  */
 #else
 #define TERMCAP		1               /* Use TERMCAP                  */
@@ -107,9 +104,14 @@ typedef struct
 #define	QSType	  void
 #define QcompType const void
 #else
+#if	defined(hpp)
+#define	QSType	  void
+#define QcompType void
+#else
 #define	QSType	  int
 #define QcompType void
 #endif
+#endif
 
 /*
  * File name separator, as a char and string
@@ -154,12 +156,10 @@ typedef struct
 #define	strrchr	rindex
 #endif	/* dyn */
 
-extern struct KBSTREE *kpadseqs;
-extern int kbseq();
-
-#if	TERMCAP
-extern struct KBSTREE *kpadseqs;
-#endif	/* TERMCAP */
+#ifdef	MOUSE
+#define	XTERM_MOUSE_ON	"\033[?1000h"	/* DECSET with parm 1000 */
+#define	XTERM_MOUSE_OFF	"\033[?1000l"	/* DECRST with parm 1000  */
+#endif
 
 #if	defined(bsd) || defined(nxt) || defined(dyn)
 #ifdef	ANSI
@@ -178,4 +178,11 @@ extern char *getcwd();
  */
 #define	MODE_READONLY	(0600)
 
+
+/*
+ * Default terminal dimensions...
+ */
+#define	NROW	24
+#define	NCOL	80
+
 #endif	/* OSDEP_H */
