@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: tcap.c,v 4.13 1994/03/17 01:00:06 mikes Exp $";
+static char rcsid[] = "$Id: tcap.c,v 4.26 1996/03/15 07:41:11 hubert Exp $";
 #endif
 /*
  * Program:	Display routines
@@ -15,28 +15,16 @@ static char rcsid[] = "$Id: tcap.c,v 4.13 1994/03/17 01:00:06 mikes Exp $";
  *
  * Please address all bugs and comments to "pine-bugs@cac.washington.edu"
  *
- * Copyright 1991-1994  University of Washington
  *
- *  Permission to use, copy, modify, and distribute this software and its
- * documentation for any purpose and without fee to the University of
- * Washington is hereby granted, provided that the above copyright notice
- * appears in all copies and that both the above copyright notice and this
- * permission notice appear in supporting documentation, and that the name
- * of the University of Washington not be used in advertising or publicity
- * pertaining to distribution of the software without specific, written
- * prior permission.  This software is made available "as is", and
- * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
- * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
- * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
- * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
- * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
- * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- * Pine and Pico are trademarks of the University of Washington.
- * No commercial use of these trademarks may be made without prior
- * written permission of the University of Washington.
+ * Pine and Pico are registered trademarks of the University of Washington.
+ * No commercial use of these trademarks may be made without prior written
+ * permission of the University of Washington.
+ * 
+ * Pine, Pico, and Pilot software and its included text are Copyright
+ * 1989-1996 by the University of Washington.
+ * 
+ * The full text of our legal notices is contained in the file called
+ * CPYRIGHT, included with this distribution.
  *
  */
 /*	tcap:	Unix V5, V7 and BS4.2 Termcap video driver
@@ -48,16 +36,14 @@ static char rcsid[] = "$Id: tcap.c,v 4.13 1994/03/17 01:00:06 mikes Exp $";
 #include	<stdio.h>
 #include        <signal.h>
 #include	"osdep.h"
+#include        "pico.h"
 #include	"estruct.h"
 #include        "edef.h"
 
 #if TERMCAP
-#include        "pico.h"
-
-#define NROW    24
-#define NCOL    80
 #define	MARGIN	8
 #define	SCRSIZ	64
+#define	MROW	2
 #define BEL     0x07
 #define ESC     0x1B
 
@@ -66,13 +52,16 @@ extern int      ttgetc();
 extern int      ttputc();
 extern int      ttflush();
 extern int      ttclose();
-extern int      tcapmove();
-extern int      tcapeeol();
-extern int      tcapeeop();
-extern int      tcapbeep();
-extern int	tcaprev();
-extern int      tcapopen();
-extern int      tcapclose();
+
+static int      tcapmove();
+static int      tcapeeol();
+static int      tcapeeop();
+static int      tcapbeep();
+static int	tcaprev();
+static int      tcapopen();
+static int      tcapclose();
+static void     setup_dflt_pico_esc_seq();
+
 extern int      tput();
 extern char     *tgoto();
 
@@ -96,16 +85,13 @@ char	*DL,			/* delete line */
 	*TI,			/* string to start termcap */
         *TE;			/* string to end termcap */
 
-char *KU, *KD, *KL, *KR;
-char *KPPU, *KPPD, *KPHOME, *KPEND;
-
-struct KBSTREE *kpadseqs = NULL;
 
 TERM term = {
         NROW-1,
         NCOL,
 	MARGIN,
 	SCRSIZ,
+	MROW,
         tcapopen,
         tcapclose,
         ttgetc,
@@ -119,6 +105,111 @@ TERM term = {
 };
 
 
+/*
+ * Add default keypad sequences to the trie.
+ */
+static void
+setup_dflt_pico_esc_seq()
+{
+    /*
+     * this is sort of a hack, but it allows us to use
+     * the function keys on pc's running telnet
+     */
+
+    /* 
+     * UW-NDC/UCS vt10[02] application mode.
+     */
+    kpinsert(&pico_kbesc, "\033OP", F1);
+    kpinsert(&pico_kbesc, "\033OQ", F2);
+    kpinsert(&pico_kbesc, "\033OR", F3);
+    kpinsert(&pico_kbesc, "\033OS", F4);
+    kpinsert(&pico_kbesc, "\033Op", F5);
+    kpinsert(&pico_kbesc, "\033Oq", F6);
+    kpinsert(&pico_kbesc, "\033Or", F7);
+    kpinsert(&pico_kbesc, "\033Os", F8);
+    kpinsert(&pico_kbesc, "\033Ot", F9);
+    kpinsert(&pico_kbesc, "\033Ou", F10);
+    kpinsert(&pico_kbesc, "\033Ov", F11);
+    kpinsert(&pico_kbesc, "\033Ow", F12);
+
+    /*
+     * DC vt100, ANSI and cursor key mode.
+     */
+    kpinsert(&pico_kbesc, "\033OA", K_PAD_UP);
+    kpinsert(&pico_kbesc, "\033OB", K_PAD_DOWN);
+    kpinsert(&pico_kbesc, "\033OC", K_PAD_RIGHT);
+    kpinsert(&pico_kbesc, "\033OD", K_PAD_LEFT);
+
+    /*
+     * special keypad functions
+     */
+    kpinsert(&pico_kbesc, "\033[4J", K_PAD_PREVPAGE);
+    kpinsert(&pico_kbesc, "\033[3J", K_PAD_NEXTPAGE);
+    kpinsert(&pico_kbesc, "\033[2J", K_PAD_HOME);
+    kpinsert(&pico_kbesc, "\033[N",  K_PAD_END);
+
+    /* 
+     * ANSI mode.
+     */
+    kpinsert(&pico_kbesc, "\033[=a", F1);
+    kpinsert(&pico_kbesc, "\033[=b", F2);
+    kpinsert(&pico_kbesc, "\033[=c", F3);
+    kpinsert(&pico_kbesc, "\033[=d", F4);
+    kpinsert(&pico_kbesc, "\033[=e", F5);
+    kpinsert(&pico_kbesc, "\033[=f", F6);
+    kpinsert(&pico_kbesc, "\033[=g", F7);
+    kpinsert(&pico_kbesc, "\033[=h", F8);
+    kpinsert(&pico_kbesc, "\033[=i", F9);
+    kpinsert(&pico_kbesc, "\033[=j", F10);
+    kpinsert(&pico_kbesc, "\033[=k", F11);
+    kpinsert(&pico_kbesc, "\033[=l", F12);
+
+    /*
+     * DEC vt100, ANSI, cursor key mode reset.
+     */
+    kpinsert(&pico_kbesc, "\033[A", K_PAD_UP);
+    kpinsert(&pico_kbesc, "\033[B", K_PAD_DOWN);
+    kpinsert(&pico_kbesc, "\033[C", K_PAD_RIGHT);
+    kpinsert(&pico_kbesc, "\033[D", K_PAD_LEFT);
+
+    /*
+     * DEC vt52 mode.
+     */
+    kpinsert(&pico_kbesc, "\033A", K_PAD_UP);
+    kpinsert(&pico_kbesc, "\033B", K_PAD_DOWN);
+    kpinsert(&pico_kbesc, "\033C", K_PAD_RIGHT);
+    kpinsert(&pico_kbesc, "\033D", K_PAD_LEFT);
+
+    /*
+     * DEC vt52 application keys, and some Zenith 19.
+     */
+    kpinsert(&pico_kbesc, "\033?r", K_PAD_DOWN);
+    kpinsert(&pico_kbesc, "\033?t", K_PAD_LEFT);
+    kpinsert(&pico_kbesc, "\033?v", K_PAD_RIGHT);
+    kpinsert(&pico_kbesc, "\033?x", K_PAD_UP);
+
+    /*
+     * Sun Console sequences.
+     */
+    kpinsert(&pico_kbesc, "\033[1",   K_SWALLOW_TIL_Z);
+    kpinsert(&pico_kbesc, "\033[215", K_SWALLOW_UP);
+    kpinsert(&pico_kbesc, "\033[217", K_SWALLOW_LEFT);
+    kpinsert(&pico_kbesc, "\033[219", K_SWALLOW_RIGHT);
+    kpinsert(&pico_kbesc, "\033[221", K_SWALLOW_DOWN);
+
+    /*
+     * Kermit App Prog Cmd, gobble until ESC \ (kermit should intercept this)
+     */
+    kpinsert(&pico_kbesc, "\033_", K_KERMIT);
+
+    /*
+     * Fake a control character.
+     */
+    kpinsert(&pico_kbesc, "\033\033", K_DOUBLE_ESC);
+}
+
+
+static int
 tcapopen()
 {
     char   *t, *p, *tgetstr();
@@ -126,8 +217,11 @@ tcapopen()
     char   *tv_stype;
     char    err_str[72];
     char   *getenv();
-
-    ttgetwinsz();
+    int     row, col;
+    char   *KU, *KD, *KL, *KR,
+	   *KPPU, *KPPD, *KPHOME, *KPEND, *KPDEL,
+	   *KF1, *KF2, *KF3, *KF4, *KF5, *KF6,
+	   *KF7, *KF8, *KF9, *KF10, *KF11, *KF12;
 
     /*
      * determine the terminal's communication speed and decide
@@ -181,6 +275,34 @@ tcapopen()
     TI = tgetstr("ti", &p);
     TE = tgetstr("te", &p);
 
+    row = tgetnum("li");
+    if(row == -1){
+	char *er;
+	int   rr;
+
+	/* tgetnum failed, try $LINES */
+	er = getenv("LINES");
+	if(er && (rr = atoi(er)) > 0)
+	  row = rr;
+    }
+    if(row >= 0)
+      row--;
+
+    col = tgetnum("co");
+    if(col == -1){
+	char *ec;
+	int   cc;
+
+	/* tgetnum failed, try $COLUMNS */
+	ec = getenv("COLUMNS");
+	if(ec && (cc = atoi(ec)) > 0)
+	  col = cc;
+    }
+
+    ttgetwinsz(&row, &col);
+    term.t_nrow = (short) row;
+    term.t_ncol = (short) col;
+
     eolexist = (CE != NULL);	/* will we be able to use clear to EOL? */
     revexist = (SO != NULL);
     if(DC == NULL && (DM == NULL || ED == NULL))
@@ -200,101 +322,90 @@ tcapopen()
 	KPPU   = tgetstr("kP", &p);
 	KPPD   = tgetstr("kN", &p);
 	KPHOME = tgetstr("kh", &p);
+	KPEND  = tgetstr("kE", &p);
+	KPDEL  = tgetstr("kD", &p);
 	KU     = tgetstr("ku", &p);
 	KD     = tgetstr("kd", &p);
 	KL     = tgetstr("kl", &p);
 	KR     = tgetstr("kr", &p);
-	if(KU != NULL && (KL != NULL && (KR != NULL && KD != NULL))){
-	    kpinsert(KU,K_PAD_UP);
-	    kpinsert(KD,K_PAD_DOWN);
-	    kpinsert(KL,K_PAD_LEFT);
-	    kpinsert(KR,K_PAD_RIGHT);
+	KF1    = tgetstr("k1", &p);
+	KF2    = tgetstr("k2", &p);
+	KF3    = tgetstr("k3", &p);
+	KF4    = tgetstr("k4", &p);
+	KF5    = tgetstr("k5", &p);
+	KF6    = tgetstr("k6", &p);
+	KF7    = tgetstr("k7", &p);
+	KF8    = tgetstr("k8", &p);
+	KF9    = tgetstr("k9", &p);
+	if((KF10 = tgetstr("k;", &p)) == NULL)
+	  KF10 = tgetstr("k0", &p);
+	KF11   = tgetstr("F1", &p);
+	KF12   = tgetstr("F2", &p);
 
-	    if(KPPU != NULL)
-	      kpinsert(KPPU,K_PAD_PREVPAGE);
-	    if(KPPD != NULL)
-	      kpinsert(KPPD,K_PAD_NEXTPAGE);
-	    if(KPHOME != NULL)
-	      kpinsert(KPHOME,K_PAD_HOME);
+#ifndef TERMCAP_WINS
+	/*
+	 * Add default keypad sequences to the trie.
+	 * Since these come first, they will override any conflicting termcap
+	 * or terminfo escape sequences defined below.  An escape sequence is
+	 * considered conflicting if one is a prefix of the other.
+	 * So, without TERMCAP_WINS, there will likely be some termcap/terminfo
+	 * escape sequences that don't work, because they conflict with default
+	 * sequences.
+	 */
+	setup_dflt_pico_esc_seq();
+#endif /* !TERMCAP_WINS */
+
+	/*
+	 * add termcap escape sequences to the trie...
+	 */
+	if(KU != NULL && (KL != NULL && (KR != NULL && KD != NULL))){
+	    kpinsert(&pico_kbesc, KU, K_PAD_UP);
+	    kpinsert(&pico_kbesc, KD, K_PAD_DOWN);
+	    kpinsert(&pico_kbesc, KL, K_PAD_LEFT);
+	    kpinsert(&pico_kbesc, KR, K_PAD_RIGHT);
 	}
+
+	if(KPPU != NULL && KPPD != NULL){
+	    kpinsert(&pico_kbesc, KPPU, K_PAD_PREVPAGE);
+	    kpinsert(&pico_kbesc, KPPD, K_PAD_NEXTPAGE);
+	}
+
+	kpinsert(&pico_kbesc, KPHOME, K_PAD_HOME);
+	kpinsert(&pico_kbesc, KPEND,  K_PAD_END);
+	kpinsert(&pico_kbesc, KPDEL,  K_PAD_DELETE);
+
+	kpinsert(&pico_kbesc, KF1,  F1);
+	kpinsert(&pico_kbesc, KF2,  F2);
+	kpinsert(&pico_kbesc, KF3,  F3);
+	kpinsert(&pico_kbesc, KF4,  F4);
+	kpinsert(&pico_kbesc, KF5,  F5);
+	kpinsert(&pico_kbesc, KF6,  F6);
+	kpinsert(&pico_kbesc, KF7,  F7);
+	kpinsert(&pico_kbesc, KF8,  F8);
+	kpinsert(&pico_kbesc, KF9,  F9);
+	kpinsert(&pico_kbesc, KF10, F10);
+	kpinsert(&pico_kbesc, KF11, F11);
+	kpinsert(&pico_kbesc, KF12, F12);
     }
 
     /*
-     * add default keypad sequences to the trie...
+     * Initialize UW-modified NCSA telnet to use it's functionkeys
      */
-    if(gmode&MDFKEY){
-	/*
-	 * Initialize UW-modified NCSA telnet to use it's functionkeys
-	 */
-	if(Pmaster == NULL){
-	    puts("\033[99h");
-	}
+    if(gmode&MDFKEY && Pmaster == NULL)
+      puts("\033[99h");
 
-	/*
-	 * this is sort of a hack, but it allows us to use
-	 * the function keys on pc's running telnet
-	 */
-
-	/* 
-	 * UW-NDC/UCS vt10[02] application mode.
-	 */
-	kpinsert("OP",F1);
-	kpinsert("OQ",F2);
-	kpinsert("OR",F3);
-	kpinsert("OS",F4);
-	kpinsert("Op",F5);
-	kpinsert("Oq",F6);
-	kpinsert("Or",F7);
-	kpinsert("Os",F8);
-	kpinsert("Ot",F9);
-	kpinsert("Ou",F10);
-	kpinsert("Ov",F11);
-	kpinsert("Ow",F12);
-
-	/*
-	 * special keypad functions
-	 */
-	kpinsert("[4J",K_PAD_PREVPAGE);
-	kpinsert("[3J",K_PAD_NEXTPAGE);
-	kpinsert("[2J",K_PAD_HOME);
-	kpinsert("[N",K_PAD_END);
-
-	/* 
-	 * ANSI mode.
-	 */
-	kpinsert("[=a",F1);
-	kpinsert("[=b",F2);
-	kpinsert("[=c",F3);
-	kpinsert("[=d",F4);
-	kpinsert("[=e",F5);
-	kpinsert("[=f",F6);
-	kpinsert("[=g",F7);
-	kpinsert("[=h",F8);
-	kpinsert("[=i",F9);
-	kpinsert("[=j",F10);
-	kpinsert("[=k",F11);
-	kpinsert("[=l",F12);
-    }
-
-    kpinsert("OA",K_PAD_UP);	/* DEC vt100, ANSI and cursor key mode. */
-    kpinsert("OB",K_PAD_DOWN);
-    kpinsert("OD",K_PAD_LEFT);
-    kpinsert("OC",K_PAD_RIGHT);
-
-    kpinsert("[A",K_PAD_UP);	/* DEC vt100, ANSI, cursor key mode reset. */
-    kpinsert("[B",K_PAD_DOWN);
-    kpinsert("[D",K_PAD_LEFT);
-    kpinsert("[C",K_PAD_RIGHT);
-
-    kpinsert("A",K_PAD_UP);	/* DEC vt52 mode. */
-    kpinsert("B",K_PAD_DOWN);
-    kpinsert("D",K_PAD_LEFT);
-    kpinsert("C",K_PAD_RIGHT);
-
-    kpinsert("[215z",K_PAD_UP); /* Sun Console sequences. */
-    kpinsert("[221z",K_PAD_DOWN);
-    kpinsert("[217z",K_PAD_LEFT);
-    kpinsert("[219z",K_PAD_RIGHT);
+#ifdef TERMCAP_WINS
+    /*
+     * Add default keypad sequences to the trie.
+     * Since these come after the termcap/terminfo escape sequences above,
+     * the termcap/info sequences will override any conflicting default
+     * escape sequences defined here.
+     * So, with TERMCAP_WINS, some of the default sequences will be missing.
+     * This means that you'd better get all of your termcap/terminfo entries
+     * correct if you define TERMCAP_WINS.
+     */
+    setup_dflt_pico_esc_seq();
+#endif /* TERMCAP_WINS */
 
     if (p >= &tcapbuf[TCAPSLEN]){
 	if(Pmaster == NULL){
@@ -313,6 +424,7 @@ tcapopen()
 }
 
 
+static int
 tcapclose()
 {
     if(!Pmaster){
@@ -323,76 +435,13 @@ tcapclose()
 	  putpad(TE);
     }
 
+    kbdestroy(pico_kbesc);		/* clean up key board sequence trie */
+    pico_kbesc = NULL;
     ttclose();
 }
 
 
 
-#define	newnode()	(struct KBSTREE *)malloc(sizeof(struct KBSTREE))
-/*
- * kbinsert - insert a keystroke escape sequence into the global search
- *	      structure.
- */
-kpinsert(kstr, kval)
-char	*kstr;
-int	kval;
-{
-    register	char	*buf;
-    register	struct KBSTREE *temp;
-    register	struct KBSTREE *trail;
-
-    if(kstr == NULL)
-      return;
-
-    temp = trail = kpadseqs;
-    if(kstr[0] == '\033')
-      buf = kstr+1;			/* can the ^[ character */ 
-    else
-      buf = kstr;
-
-    for(;;) {
-	if(temp == NULL){
-	    temp = newnode();
-	    temp->value = *buf;
-	    temp->func = 0;
-	    temp->left = NULL;
-	    temp->down = NULL;
-	    if(kpadseqs == NULL)
-	      kpadseqs = temp;
-	    else
-	      trail->down = temp;
-	}
-	else{				/* first entry */
-	    while((temp != NULL) && (temp->value != *buf)){
-		trail = temp;
-		temp = temp->left;
-	    }
-
-	    if(temp == NULL){   /* add new val */
-		temp = newnode();
-		temp->value = *buf;
-		temp->func = 0;
-		temp->left = NULL;
-		temp->down = NULL;
-		trail->left = temp;
-	    }
-	}
-
-	if (*(++buf) == '\0'){
-	    break;
-	}
-	else{
-	    trail = temp;
-	    temp = temp->down;
-	}
-    }
-    
-    if(temp != NULL)
-      temp->func = kval;
-}
-
-
-
 /*
  * tcapinsert - insert a character at the current character position.
  *              IC takes precedence.
@@ -439,7 +488,7 @@ register int n;
     register int i;
 
     if(CS != NULL){
-	putpad(tgoto(CS, term.t_nrow - 3, row));
+	putpad(tgoto(CS, term.t_nrow - (term.t_mrow+1), row));
 	tcapmove(row, 0);
 	for(i = 0; i < n; i++)
 	  putpad( (SR != NULL && *SR != '\0') ? SR : "\n" );
@@ -451,7 +500,7 @@ register int n;
 	 * this code causes a jiggly motion of the keymenu when scrolling
 	 */
 	for(i = 0; i < n; i++){
-	    tcapmove(term.t_nrow - 3, 0);
+	    tcapmove(term.t_nrow - (term.t_mrow+1), 0);
 	    putpad(DL);
 	    tcapmove(row, 0);
 	    putpad(AL);
@@ -460,7 +509,7 @@ register int n;
 	/*
 	 * this code causes a sweeping motion up and down the display
 	 */
-	tcapmove(term.t_nrow - 2 - n, 0);
+	tcapmove(term.t_nrow - term.t_mrow - n, 0);
 	for(i = 0; i < n; i++)
 	  putpad(DL);
 	tcapmove(row, 0);
@@ -483,9 +532,9 @@ register int n;
     register int i;
 
     if(CS != NULL){
-	putpad(tgoto(CS, term.t_nrow - 3, row));
+	putpad(tgoto(CS, term.t_nrow - (term.t_mrow+1), row));
 	/* setting scrolling region moves cursor to home */
-	tcapmove(term.t_nrow-3, 0);
+	tcapmove(term.t_nrow-(term.t_mrow+1), 0);
 	for(i = 0;i < n; i++)
 	  putpad((SF == NULL || SF[0] == '\0') ? "\n" : SF);
 	putpad(tgoto(CS, term.t_nrow, 0));
@@ -495,7 +544,7 @@ register int n;
 	for(i = 0; i < n; i++){
 	    tcapmove(row, 0);
 	    putpad(DL);
-	    tcapmove(term.t_nrow - 3, 0);
+	    tcapmove(term.t_nrow - (term.t_mrow+1), 0);
 	    putpad(AL);
 	}
 #ifdef  NOWIGGLYLINES
@@ -503,7 +552,7 @@ register int n;
 	tcapmove(row, 0);
 	for(i = 0; i < n; i++)
 	  putpad(DL);
-	tcapmove(term.t_nrow - 2 - n, 0);
+	tcapmove(term.t_nrow - term.t_mrow - n, 0);
 	for(i = 0;i < n; i++)
 	  putpad(AL);
 #endif
@@ -542,6 +591,7 @@ o_delete()
 }
 
 
+static int
 tcapmove(row, col)
 register int row, col;
 {
@@ -549,18 +599,21 @@ register int row, col;
 }
 
 
+static int
 tcapeeol()
 {
     putpad(CE);
 }
 
 
+static int
 tcapeeop()
 {
     putpad(CL);
 }
 
 
+static int
 tcaprev(state)		/* change reverse video status */
 int state;	        /* FALSE = normal video, TRUE = reverse video */
 {
@@ -580,6 +633,7 @@ int state;	        /* FALSE = normal video, TRUE = reverse video */
 }
 
 
+static int
 tcapbeep()
 {
     ttputc(BEL);
