@@ -1,5 +1,5 @@
 #if	!defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: word.c,v 4.3 1993/11/08 20:22:47 mikes Exp $";
+static char rcsid[] = "$Id: word.c,v 4.20 1996/03/15 07:41:11 hubert Exp $";
 #endif
 /*
  * Program:	Word at a time routines
@@ -15,28 +15,16 @@ static char rcsid[] = "$Id: word.c,v 4.3 1993/11/08 20:22:47 mikes Exp $";
  *
  * Please address all bugs and comments to "pine-bugs@cac.washington.edu"
  *
- * Copyright 1991-1993  University of Washington
  *
- *  Permission to use, copy, modify, and distribute this software and its
- * documentation for any purpose and without fee to the University of
- * Washington is hereby granted, provided that the above copyright notice
- * appears in all copies and that both the above copyright notice and this
- * permission notice appear in supporting documentation, and that the name
- * of the University of Washington not be used in advertising or publicity
- * pertaining to distribution of the software without specific, written
- * prior permission.  This software is made available "as is", and
- * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
- * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
- * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
- * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
- * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
- * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
- * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- * Pine and Pico are trademarks of the University of Washington.
- * No commercial use of these trademarks may be made without prior
- * written permission of the University of Washington.
+ * Pine and Pico are registered trademarks of the University of Washington.
+ * No commercial use of these trademarks may be made without prior written
+ * permission of the University of Washington.
+ * 
+ * Pine, Pico, and Pilot software and its included text are Copyright
+ * 1989-1996 by the University of Washington.
+ * 
+ * The full text of our legal notices is contained in the file called
+ * CPYRIGHT, included with this distribution.
  *
  */
 /*
@@ -46,11 +34,11 @@ static char rcsid[] = "$Id: word.c,v 4.3 1993/11/08 20:22:47 mikes Exp $";
  */
 
 #include        <stdio.h>
-#include        "estruct.h"
+#include	"osdep.h"
 #include        "pico.h"
+#include        "estruct.h"
 #include        <ctype.h>
 #include	"edef.h"
-#include	"osdep.h"
 
 
 /* Word wrap on n-spaces. Back-over whatever precedes the point on the current
@@ -62,50 +50,35 @@ static char rcsid[] = "$Id: word.c,v 4.3 1993/11/08 20:22:47 mikes Exp $";
  */
 wrapword()
 {
-    register int   cnt = 0;	/* size of word wrapped to next line */
-    register int   bp;
+    register int cnt;			/* size of word wrapped to next line */
+    register int bp;			/* index to wrap on */
+    register int first = -1;
+    register int i;
 
     if(curwp->w_doto <= 0)		/* no line to wrap? */
       return(FALSE);
 
-    /* 
-     * back up until we aren't in a word,
-     * and make sure there is a break in the line
-     */
-    bp = llength(curwp->w_dotp);	/* start at end of line */
-    do{
-	if(--bp <= 0)			/* no place to break the line! */
-	  return(FALSE);
-	
-	switch(cnt){
-	  case 0:			/* break BEFORE fillcol */
-	    if(bp > fillcol)
-	      break;
-	    else
-	      cnt++;
-
-	  case 1:			/* find first breakable word */
-	    if(isspace(lgetc(curwp->w_dotp, bp).c))
-	      break;
-	    else
-	      cnt++;
-
-	  case 2:			/* and break in front of it */
-	    if(!isspace(lgetc(curwp->w_dotp, bp).c))
-	      break;
-	    else
-	      cnt = 10;
-
-	  default:
-	    break;
+    for(bp = cnt = i = 0; cnt < llength(curwp->w_dotp) && !bp; cnt++, i++){
+	if(isspace(lgetc(curwp->w_dotp, cnt).c)){
+	    first = 0;
+	    if(lgetc(curwp->w_dotp, cnt).c == TAB)
+	      while(i+1 & 0x07)
+		i++;
 	}
-    }
-    while(cnt != 10);
+	else if(!first)
+	  first = cnt;
 
-    /* bp now points to the last character to remain on this line! */
-    cnt = curwp->w_doto - ++bp;
+	if(first > 0 && i >= fillcol)
+	  bp = first;
+    }
+
+    if(!bp)
+      return(FALSE);
+
+    /* bp now points to the first character of the next line */
+    cnt = curwp->w_doto - bp;
     curwp->w_doto = bp;
-    
+
     if(!lnewline())			/* break the line */
       return(FALSE);
 
@@ -113,7 +86,8 @@ wrapword()
      * if there's a line below, it doesn't start with whitespace 
      * and there's room for this line...
      */
-    if(lforw(curwp->w_dotp) != curbp->b_linep 
+    if(!(curbp->b_flag & BFWRAPOPEN)
+       && lforw(curwp->w_dotp) != curbp->b_linep 
        && llength(lforw(curwp->w_dotp)) 
        && !isspace(lgetc(lforw(curwp->w_dotp), 0).c)
        && (llength(curwp->w_dotp) + llength(lforw(curwp->w_dotp)) < fillcol)){
@@ -125,7 +99,8 @@ wrapword()
 	gotobol(0, 1);
     }
 
-    if(!forwchar(0, cnt < 0 ? cnt-1 : cnt)) /* restore dot (account for NL) */
+    curbp->b_flag &= ~BFWRAPOPEN;           /* don't open new line next wrap */
+    if(!forwchar(0, cnt < 0 ? cnt-1 : cnt)) /* restore dot (account for NL)  */
       return(FALSE);
 
     return(TRUE);
@@ -306,7 +281,7 @@ capword(f, n)
  */
 delfword(f, n)
 {
-        register int    size;
+        register long   size;
         register LINE   *dotp;
         register int    doto;
 
@@ -316,7 +291,7 @@ delfword(f, n)
                 return (FALSE);
         dotp = curwp->w_dotp;
         doto = curwp->w_doto;
-        size = 0;
+        size = 0L;
         while (n--) {
 #if	NFWORD
 		while (inword() != FALSE) {
@@ -350,7 +325,7 @@ delfword(f, n)
  */
 delbword(f, n)
 {
-        register int    size;
+        register long   size;
 
 	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
 		return(rdonly());	/* we are in read only mode	*/
@@ -358,7 +333,7 @@ delbword(f, n)
                 return (FALSE);
         if (backchar(FALSE, 1) == FALSE)
                 return (FALSE);
-        size = 0;
+        size = 0L;
         while (n--) {
                 while (inword() == FALSE) {
                         if (backchar(FALSE, 1) == FALSE)
@@ -383,139 +358,268 @@ delbword(f, n)
  */
 inword()
 {
-        register int    c;
-
-        if (curwp->w_doto == llength(curwp->w_dotp))
-                return (FALSE);
-        c = lgetc(curwp->w_dotp, curwp->w_doto).c;
-
-        if (c>='a' && c<='z')
-                return (TRUE);
-        if (c>='A' && c<='Z')
-                return (TRUE);
-        if (c>='0' && c<='9')
-                return (TRUE);
-        return (FALSE);
+    return(curwp->w_doto < llength(curwp->w_dotp)
+	   && isalnum(lgetc(curwp->w_dotp, curwp->w_doto).c));
 }
 
+
 fillpara(f, n)	/* Fill the current paragraph according to the current
 		   fill column						*/
 
 int f, n;	/* deFault flag and Numeric argument */
 
 {
-	register int c;			/* current char durring scan	*/
-	register int wordlen;		/* length of current word	*/
-	register int clength;		/* position on line during fill	*/
-	register int i;			/* index during word copy	*/
-	register int newlength;		/* tentative new line length	*/
-	register int eopflag;		/* Are we at the End-Of-Paragraph? */
-	register int firstflag;		/* first word? (needs no space)	*/
-	register LINE *eopline;		/* pointer to line just past EOP */
-	register int dotflag;		/* was the last char a period?	*/
-	char wbuf[NSTRING];		/* buffer for current word	*/
+    register int c;			/* current char durring scan	*/
+    register int wordlen;		/* length of current word	*/
+    register int clength;		/* position on line during fill	*/
+    register int i;			/* index during word copy	*/
+    register int newlength;		/* tentative new line length	*/
+    register int eopflag;		/* Are we at the End-Of-Paragraph? */
+    register int firstflag;		/* first word? (needs no space)	*/
+    register LINE *eopline;		/* pointer to line just past EOP */
+    register int dotflag;		/* was the last char a period?	*/
+    char wbuf[NSTRING];			/* buffer for current word	*/
+    int  first_word_of_paragraph;
+    int  end_of_input_line;
+    char *quote = NULL;
+    int   quote_len;
+    LINE *bqline, *eqline, *qline, *lp;
 
+    if(curbp->b_mode&MDVIEW){		/* don't allow this command if	*/
+	return(rdonly());		/* we are in read only mode	*/
+    }
+    else if (fillcol == 0) {		/* no fill column set */
+	mlwrite("No fill column set");
+	return(FALSE);
+    }
 
-	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
-		return(rdonly());	/* we are in read only mode	*/
-	if (fillcol == 0) {	/* no fill column set */
-		mlwrite("No fill column set");
-		return(FALSE);
+    qline = curwp->w_dotp;		/* remember where we started */
+
+    /* record the pointer to the line just past the EOP */
+    if(gotoeop(FALSE, 1) == FALSE)
+      return(FALSE);
+
+    eopline = lforw(curwp->w_dotp);
+
+    /* and back to the beginning of the paragraph */
+    gotobop(FALSE, 1);
+
+    /* if we're the composer, and we were given a quoting string,
+     * and it starts the line we're on, do some special magic...
+     */
+    if(Pmaster && Pmaster->quote_str
+       && (quote_len = i = strlen(Pmaster->quote_str))
+       && i <= llength(curwp->w_dotp)){
+	while(--i >= 0 && Pmaster->quote_str[i] == lgetc(curwp->w_dotp, i).c)
+	  ;
+
+	if(i < 0){				/* bingo! */
+#ifdef	MAYBELATER
+	    switch(mlyesno("Format quoted text", TRUE)){
+	      case FALSE :
+		gotoeop(FALSE, 1);
+		forwchar(FALSE, 1);
+	      case ABORT :
+		emlwrite("Text not Justified", NULL);
+		return(TRUE);
+
+	      default :
+		emlwrite("Justifying quoted text", NULL);
+	    }
+#endif
+
+	    /*
+	     * adjust fillcol and bracket the quoted text with special
+	     * blank lines
+	     */
+	    quote = Pmaster->quote_str;
+	    fillcol -= quote_len;
+
+	    i = curwp->w_doto;
+	    gotobol(FALSE, 1);
+	    lnewline();				/* insert special blank line */
+	    backchar(FALSE, 1);			/* back onto it */
+	    bqline = curwp->w_dotp;		/* mark start of quoted text */
+	    curwp->w_dotp = eopline;
+	    gotoeol(FALSE, 1);
+	    lnewline();
+	    eqline = curwp->w_dotp;		/* mark end of quoted text */
+	    curwp->w_dotp = lforw(bqline);	/* restore dot */
+	    curwp->w_doto = i;
 	}
+    }
 
-	/* record the pointer to the line just past the EOP */
+    /* let yank() know that it may be restoring a paragraph */
+    thisflag |= CFFILL;
+    
+    if(Pmaster == NULL)
+      sgarbk = TRUE;
+    
+    curwp->w_flag |= WFMODE;
+
+    /* save text to be justified to the cut buffer */
+    fdelete();
+    for(lp = curwp->w_dotp;
+	lp != curbp->b_linep && lp != eopline;
+	lp = lforw(lp)){
+	for(i=(lp == curwp->w_dotp) ? curwp->w_doto : 0; i < llength(lp); i++)
+	  finsert(lgetc(lp, i).c);
+
+	finsert('\n');
+    }
+
+    if(quote){
+	/* remove quoting from whole shebang */
+	curwp->w_dotp = lforw(bqline);
+	curwp->w_doto = 0;
+	do
+	  forwdel(FALSE, (quote_len > llength(curwp->w_dotp))
+			   ? llength(curwp->w_dotp) : quote_len);
+	while((curwp->w_dotp = lforw(curwp->w_dotp)) != lback(eqline));
+
+	/* got to the line we started on */
+	curwp->w_dotp = qline;
+	curwp->w_doto = 0;
 	gotoeop(FALSE, 1);
-	eopline = lforw(curwp->w_dotp);
+	eopline = lforw(curwp->w_dotp);	/* eop within quoted text */
+	/* make sure new eop isn't beyond special token line */
+	for(lp = curwp->w_dotp;
+	    lp != curbp->b_linep && lp != lback(eqline);
+	    lp = lforw(lp))
+	  ;
 
-	/* and back to the beginning of the paragraph */
-	gotobop(FALSE, 1);
-
-	/* let yank() know that it may be restoring a paragraph */
-	thisflag |= CFFILL;
-
-	if(Pmaster == NULL)
-	  sgarbk = TRUE;
-
-        curwp->w_flag |= WFMODE;
-	kdelete();
-
-	/* initialize various info */
-	clength = curwp->w_doto;
-	if (clength && curwp->w_dotp->l_text[0].c == TAB)
-		clength = 8;
-	wordlen = 0;
-	dotflag = FALSE;
-
-	/* scan through lines, filling words */
-	firstflag = TRUE;
-	eopflag = FALSE;
-
-	while (!eopflag) {
-		/* get the next character in the paragraph */
-		if (curwp->w_doto == llength(curwp->w_dotp)) {
-
-			c = ' ';
-			if (lforw(curwp->w_dotp) == eopline)
-				eopflag = TRUE;
-			kinsert('\n');
-		} else {
-			c = lgetc(curwp->w_dotp, curwp->w_doto).c;
-			kinsert(c);
-		}
-
-		/* and then delete it */
-		ldelete(1, FALSE);
-
-		/* if not a separator, just add it in */
-		if (c != ' ' && c != '	') {
-			/* 
-			 * don't want to limit ourselves to only '.'
-			 */
-			dotflag = (int)strchr(".?!:;\"", c);	/* dot ? */
-			if (wordlen < NSTRING - 1)
-				wbuf[wordlen++] = c;
-		} else if (wordlen) {
-			/* at a word break with a word waiting */
-			/* calculate tantitive new length with word added */
-			newlength = clength + 1 + wordlen;
-			if (newlength <= fillcol) {
-				/* add word to current line */
-				if (!firstflag) {
-					linsert(1, ' '); /* the space */
-					++clength;
-				}
-				firstflag = FALSE;
-			} else {
-				/* start a new line */
-				lnewline();
-				clength = 0;
-			}
-
-			/* and add the word in in either case */
-			for (i=0; i<wordlen; i++) {
-				linsert(1, wbuf[i]);
-				++clength;
-			}
-
-			/*  Strategy:  Handle 3 cases:
-			 *     1. if . at end of line put extra space after it
-			 *     2. if . and only 1 space, leave only one space 
-			 *     3. if . and more than 1 space, leave 2 spaces
-			 *
-			 *  So, we know the current c is a space. if then 
-			 *  is no next c or the next c is a ' ' then we 
-			 *  need to insert a space else don't do it.
-			 */
-			if (dotflag &&
-		           ((curwp->w_doto == llength(curwp->w_dotp)) || 
-			   (' ' == lgetc(curwp->w_dotp, curwp->w_doto).c))){
-				linsert(1, ' ');
-				++clength;
-			}
-			wordlen = 0;
-		}
+	/* no quoted text to justify, just clean up and return */
+	if(lp == curbp->b_linep){
+	    fillcol += quote_len;	/* restore fillcol */
+	    curwp->w_dotp = bqline;	/* blast leading special blank line */
+	    gotobol(FALSE, 1);
+	    forwdel(FALSE, 1);
+	    do{				/* restore quoting */
+		curwp->w_doto = 0;
+		for(i = 0; i < quote_len; i++)
+		  linsert(1, quote[i]);
+	    }
+	    while((curwp->w_dotp = lforw(curwp->w_dotp)) != lback(eqline));
+	    curwp->w_dotp = lback(eqline);
+	    gotoeol(FALSE, 1);
+	    forwdel(FALSE, 1);		/* blast trailing special blank line */
+	    gotobol(FALSE, 1);
+	    fdelete();			/* blast pre-justify text */
+	    return(TRUE);
 	}
 
-	/* and add a last newline for the end of our new paragraph */
-	lnewline();
+	curwp->w_dotp = lback(eopline);
+	curwp->w_doto = 0;
+	gotobop(FALSE, 1);
+    }
+    
+    /* initialize various info */
+    clength = curwp->w_doto;
+    if (clength && curwp->w_dotp->l_text[0].c == TAB)
+      clength = 8;
+    wordlen = 0;
+    dotflag = FALSE;
+    
+    /* scan through lines, filling words */
+    firstflag = TRUE;
+    eopflag = FALSE;
+    first_word_of_paragraph = TRUE;
+    
+    while (!eopflag) {
+	/* get the next character in the paragraph */
+	if (curwp->w_doto == llength(curwp->w_dotp)) {
+	    c = ' ';
+	    end_of_input_line = TRUE;
+	    if (lforw(curwp->w_dotp) == eopline)
+	      eopflag = TRUE;
+	}
+	else{
+	    c = lgetc(curwp->w_dotp, curwp->w_doto).c;
+	    end_of_input_line = FALSE;
+	}
+
+	/* and then delete it */
+	ldelete(1L, FALSE);
+
+	/* if not a separator, just add it in */
+	if (c != ' ' && c != TAB) {
+	    /* 
+	     * don't want to limit ourselves to only '.'
+	     */
+	    dotflag = (int)strchr(".?!:;\")", c); /* dot ? */
+	    if (wordlen < NSTRING - 1)
+	      wbuf[wordlen++] = c;
+	} else if (wordlen) {
+	    /* at a word break with a word waiting */
+	    /* calculate tentitive new length with word added */
+	    newlength = clength + 1 + wordlen;
+	    if (newlength <= fillcol) {
+		/* add word to current line */
+		if (!firstflag) {
+		    linsert(1, ' ');	/* the space */
+		    ++clength;
+		}
+		firstflag = FALSE;
+	    } else {
+		/* start a new line */
+		if(!first_word_of_paragraph)
+		  lnewline();
+		clength = 0;
+	    }
+
+	    first_word_of_paragraph = FALSE;
+
+	    /* and add the word in in either case */
+	    for (i=0; i<wordlen; i++) {
+		linsert(1, wbuf[i]);
+		++clength;
+	    }
+
+	    /*  Strategy:  Handle 3 cases:
+	     *     1. if . at end of line put extra space after it
+	     *     2. if . and only 1 space, leave only one space 
+	     *     3. if . and more than 1 space, leave 2 spaces
+	     *
+	     *  So, we know the current c is a space. if then 
+	     *  is no next c or the next c is a ' ' then we 
+	     *  need to insert a space else don't do it.
+	     */
+	    if(dotflag && (end_of_input_line
+			   || (' ' == lgetc(curwp->w_dotp, curwp->w_doto).c))){
+		linsert(1, ' ');
+		++clength;
+	    }
+
+	    wordlen = 0;
+	}
+    }
+    
+    /* and add a last newline for the end of our new paragraph */
+    lnewline();
+    
+    if(quote){				/* replace quoting strings */
+	fillcol += quote_len;
+	/* remember the current dot. note: ldelete at eob special, handle it */
+	eopline = (lforw(curwp->w_dotp) == curbp->b_linep)
+		     ? curbp->b_linep : curwp->w_dotp;
+	curwp->w_dotp = bqline;
+	forwdel(FALSE, 1);		/* blast special blank line */
+	do{
+	    curwp->w_doto = 0;		/* fill quote strings back in */
+	    for(i = 0; i < quote_len; i++)
+	      linsert(1, quote[i]);
+	}
+	while((curwp->w_dotp = lforw(curwp->w_dotp)) != lback(eqline));
+
+	curwp->w_dotp = lback(eqline);	/* blast ending special blank line */
+	gotobol(FALSE, 1);
+	forwdel(FALSE, 1);
+
+	/*
+	 * Restore cursor to eop w/in quote.  Note special handling 
+	 * when end of quoted text and end of buffer are the same...
+	 */
+	curwp->w_dotp = eopline;
+	curwp->w_doto = (quote_len > llength(curwp->w_dotp)) ? 0 : quote_len;
+    }
 }
