@@ -1,5 +1,5 @@
 /*----------------------------------------------------------------------
-  $Id: adrbklib.h,v 4.33 1994/10/04 21:46:17 hubert Exp $
+  $Id: adrbklib.h,v 4.56 1996/03/18 18:01:57 hubert Exp $
 
             T H E    P I N E    M A I L   S Y S T E M
 
@@ -14,28 +14,17 @@
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1994  University of Washington
 
-    Permission to use, copy, modify, and distribute this software and its
-   documentation for any purpose and without fee to the University of
-   Washington is hereby granted, provided that the above copyright notice
-   appears in all copies and that both the above copyright notice and this
-   permission notice appear in supporting documentation, and that the name
-   of the University of Washington not be used in advertising or publicity
-   pertaining to distribution of the software without specific, written
-   prior permission.  This software is made available "as is", and
-   THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
-   WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
-   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
-   NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
-   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
-   LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
-   (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
-   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
-  
-   Pine and Pico are trademarks of the University of Washington.
-   No commercial use of these trademarks may be made without prior
-   written permission of the University of Washington.
+   Pine and Pico are registered trademarks of the University of Washington.
+   No commercial use of these trademarks may be made without prior written
+   permission of the University of Washington.
+
+   Pine, Pico, and Pilot software and its included text are Copyright
+   1989-1996 by the University of Washington.
+
+   The full text of our legal notices is contained in the file called
+   CPYRIGHT, included with this distribution.
+
 
    Pine is in part based on The Elm Mail System:
     ***********************************************************************
@@ -52,7 +41,7 @@
 #define _ADRBKLIB_INCLUDED
 
 /*
- *  Some notes
+ *  Some notes:
  *
  * There is considerable indirection going on to get to an address book
  * entry.  The idea is that we don't want to have to read in the entire
@@ -77,7 +66,7 @@
  * called.  That looks in the appropriate hash bucket for 13 and runs
  * through the list looking for a match on element 13.  If it doesn't find
  * it, it reads the entry from the addrbook.lu file as above.  That entry
- * can be found easily because it is at a calculatable offset from the
+ * can be found easily because it is at a calculable offset from the
  * start of the addrbook.lu file.  Then adrbk_init_ae() is called to parse
  * the entry and fill in an ae, which is pointed to by the cached EntryRef.
  * That's how you look up a particular entry *number* in the addrbook.  That
@@ -100,15 +89,15 @@
  * to be very fast, so is done through another hashtable.  This table is
  * also precomputed and stored in the addrbook.lu file.  When the addrbook
  * is first opened, that hash table is read into memory.  Each entry in the
- * hashtable is an unsigned short element number, which points to the head
+ * hashtable is a adrbk_cntr_t element number, which points to the head
  * of the list of entries with the same hash value.  That list is what
  * uses the next_nick variable in the EntryRef.  The hashtable is not very
- * large.  Just two bytes per entry and size of hashtable entries.  The
+ * large.  Just four (or 2) bytes per entry and size of hashtable entries.  The
  * size of the table varies with the size of the addrbook, and is set by
  * the function hashtable_size().  It is also stored in the addrbook.lu
  * file so need not be the value returned by hashtable_size.  That hashtable
- * is kept open the whole time Pine is running, once it has been opened.
- * To look up an entry for nickname "nick", you compute the hash value (ab_hash)
+ * is kept open the whole time Pine is running, once it has been opened.  To
+ * look up an entry for nickname "nick", you compute the hash value (ab_hash)
  * of "nick" and the uid (ab_uid) of "nick".  Look in the hash array at
  * element number hash_value and find a pointer (an index into the EntryRef
  * array).  Now you would get that EntryRef just like above, since you
@@ -122,24 +111,24 @@
  *
  * Some sizes:
  *
- *  AdrHash is an array of shorts of length hashtable_size.  At most 40,000.
+ *  AdrHash is an array of adrbk_cntr_t of length hashtable_size.
+ *  This size is usually set by the hashtable_size function.
  *  There are two of those, one for nicknames and one for addresses.
  *  These are kept open the whole time the program is being run (once opened).
  *  Those are also stored on disk in addrbook.lu.  They take up more
- *  space there since the unsigned shorts are written out in ASCII for
- *  portability.
+ *  space there since they are written out in ASCII for portability.
  *
- *  An EntryRef cache element is size 16.  The size of this cache is usually
+ *  An EntryRef cache element is size 16-20.  The size of this cache is usually
  *  kept fairly small (200) but sorting cries out for it to be larger.
  *  When sorting in Unix, it is set equal to the "count" of the addrbook,
  *  smaller for DOS.  Each valid EntryRef cache element points to an
- *  EntryRef in memory.  Each of those is 20 fixed bytes.  Part of that is
+ *  EntryRef in memory.  Each of those is 20-24 fixed bytes.  Part of that is
  *  a pointer to a possible cached AdrBk_Entry.  An AdrBk_Entry is 22 bytes
  *  of fixed space, but it also has pointers to variable amounts of memory
  *  for nickname, fullname, addresses, fcc, and comments.  A typical
  *  entry probably uses about 50 bytes of that extra space, so altogether,
  *  a cached EntryRef with a fully-filled in ae takes around 100 bytes.
- *  If you had a 30,000 entry addrbook, setting the cache size to count
+ *  If you had a 30,000 entry addrbook, setting the cache size to "count"
  *  for sorting takes up about 3,000,000 bytes of memory.
  *
  * That's the story for adrbklib.c.  There is also some allocation happening
@@ -162,47 +151,88 @@
  * of reading through all the entries to get there.  If you edit an entry
  * so that it sorts into a new location, you warp to that new location to
  * save processing all of the entries in between.
+ *
+ *
+ * Notes about RFC1522 encoding:
+ *
+ * If the fullname field contains other than US-ASCII characters, it is
+ * encoded using the rules of RFC1522 or its successor.  The value actually
+ * stored in the file is encoded, even if it matches the character set of
+ * the user.  This is so it is easy to pass the entry around or to change
+ * character sets without invalidating entries in the address book.  When
+ * a fullname is displayed, it is first decoded.  If the fullname field is
+ * encoded as being from a character set other than the user's character
+ * set, that will be retained until the user edits the field.  Then it will
+ * change over to the user's character set.  The comment field works in
+ * the same way, as do the "phrase" fields of any addresses.  On outgoing
+ * mail, the correct character set will be retained if you use ComposeTo
+ * from the address book screen.  However, if you use a nickname from the
+ * composer or ^T from the composer, the character set will be lost if it
+ * is different from the user's character set.
  */
 
-/* #define HUGE_ADDRBOOKS */
+#define NFIELDS 11 /* one more than num of data fields in addrbook entry */
+
+/*
+ * Disk file takes up more space when HUGE is defined, but we think it is
+ * not enough more to offset the convenience of no limits.  So it is always
+ * defined.  Should still work if you want to undefine it.
+ */
+#define HUGE_ADDRBOOKS
+
+/*
+ * The type a_c_arg_t is a little confusing.  It's the type we use in
+ * place of adrbk_cntr_t when we want to pass an adrbk_cntr_t in an argument.
+ * We were running into problems with the integral promotion of adrbk_cntr_t
+ * args.  A_c_arg_t has to be large enough to hold a promoted adrbk_cntr_t.
+ * So, if adrbk_cntr_t is unsigned short, then a_c_arg_t needs to be int if
+ * int is larger than short, or unsigned int if short is same size as int.
+ * Since usign16_t always fits in a short, a_c_arg_t of unsigned int should
+ * always work for !HUGE.  For HUGE, usign32_t will be either an unsigned int
+ * or an unsigned long.  If it is an unsigned long, then a_c_arg_t better be
+ * an unsigned long, too.  If it is an unsigned int, then a_c_arg_t could
+ * be an unsigned int, too.  However, if we just make it unsigned long, then
+ * it will be the same in all cases and big enough in all cases.
+ *
+ * In the HUGE case, we could use usign32_t for the a_c_arg_t typedef.
+ * There is no actual advantage to be gained, though.  The only place it
+ * would make a difference is on machines where an int is 32 bits and a
+ * long is 64 bits, in which case the 64-bit long is probably the more
+ * efficient size to use anyway.
+ */
 
 #ifdef HUGE_ADDRBOOKS
 
-/*
- * Assume long >= 32 bits and use only 32 bits for portability of addrbook.
- */
-typedef unsigned long adrbk_cntr_t;  /* addrbook counter type                */
+typedef usign32_t adrbk_cntr_t;  /* addrbook counter type                */
 typedef unsigned long a_c_arg_t;     /* type of arg passed for adrbk_cntr_t  */
+#define NO_NEXT ((adrbk_cntr_t)-1)
 #define MAX_ADRBK_SIZE (2000000000L) /* leave room for extra display lines   */
 #define MAX_HASHTABLE_SIZE 150000
 
 # else /* !HUGE_ADDRBOOKS */
 
-/*
- * Assume short >= 16 bits and use only 16 bits for portability of addrbook.
- */
-typedef unsigned short adrbk_cntr_t; /* addrbook counter type                */
+typedef usign16_t adrbk_cntr_t; /* addrbook counter type                */
 typedef unsigned a_c_arg_t;          /* type of arg passed for addrbk_cntr_t */
+#define NO_NEXT ((adrbk_cntr_t)-1)
 #define MAX_ADRBK_SIZE ((long)(NO_NEXT - 2))
 #define MAX_HASHTABLE_SIZE 60000
 
 # endif /* !HUGE_ADDRBOOKS */
 
 /*
- * This value is reserved to mean that there is no next address, or that
+ * The value NO_NEXT is reserved to mean that there is no next address, or that
  * there is no address number to return.  This is similar to getc returning
  * -1 when there is no char to get, but since we've defined this to be
  * unsigned we need to reserve one of the valid values for this purpose.
  * With current implementation it needs to be all 1's, so memset initialization
  * will work correctly.
  */
-#define NO_NEXT ((adrbk_cntr_t)-1)
 
 typedef long adrbk_uid_t;   /* the UID of a name or address */
 
 typedef enum {ReadOnly, ReadWrite, NoAccess, NoExists} AccessType;
 typedef enum {NotSet, Single, List} Tag;
-typedef enum {Normal, Delete, Lock, Unlock, StaticOK} Handling;
+typedef enum {Normal, Delete, SaveDelete, Lock, Unlock} Handling;
 
 /* This is what is actually used by the routines that manipulate things */
 typedef struct adrbk_entry {
@@ -256,10 +286,17 @@ typedef struct width_stuff {
     int max_nickname_width;
     int max_fullname_width;
     int max_addrfield_width;
+    int max_fccfield_width;
     int third_biggest_fullname_width;
     int third_biggest_addrfield_width;
+    int third_biggest_fccfield_width;
 } WIDTH_INFO_S;
 
+typedef struct expanded_list {
+    adrbk_cntr_t          ent;
+    struct expanded_list *next;
+} EXPANDED_S;
+
 typedef struct adrbk {
     char         *orig_filename;       /* passed in filename                 */
     char         *filename;            /* addrbook filename                  */
@@ -279,17 +316,35 @@ typedef struct adrbk {
     ER_CACHE_ELEM_S **tail_cache_elem;
     int            *n_ae_cached_in_this_bucket;
     long            nominal_max_cached;
-    int             er_hashsize;
+    adrbk_cntr_t    er_hashsize;
     WIDTH_INFO_S  widths;              /* helps addrbook.c format columns    */
+    long          deleted_cnt;         /* how many #DELETED entries in abook */
     int           sort_rule;
+    EXPANDED_S   *exp;                 /* this is for addrbook.c to use.  A
+			       list of expanded list entry nums is kept here */
+    EXPANDED_S   *checks;              /* this is for addrbook.c to use.  A
+			       list of checked entry nums is kept here */
 } AdrBk;
 
+#define ONE_HUNDRED_DAYS (60L * 60L * 24L * 100L)
+/*
+ * When address book entries are deleted, they are left in the file
+ * with the nickname prepended with a string like #DELETED-96/01/25#, 
+ * which stands for year 96, month 1, day 25 of the month.  When one of
+ * these entries is more than ABOOK_DELETED_EXPIRE_TIME seconds old,
+ * then it will be totally removed from the address book the next time
+ * and adrbk_write() is done.  This is for emergencies where somebody
+ * deletes something from their address book and would like to get it
+ * back.  You get it back by editing the nickname field manually to remove
+ * the extra 18 characters off the front.
+ */
+#define ABOOK_DELETED_EXPIRE_TIME   ONE_HUNDRED_DAYS
+
 
 /*
  * There are no restrictions on the length of any of the fields, except that
  * there are some restrictions in the current input routines.
- * There is a restriction on the number of addresses in a list of MAXLIST (1000)
- * There can be no more than 65534 entries in a single addrbook.
+ * There can be no more than 65534 entries (unless HUGE) in a single addrbook.
  */
 
 /*
@@ -334,9 +389,9 @@ typedef struct adrbk {
  *            <SPACE> two character version string
  *            <SPACE> hash table size "\n" integer padded on left with spaces
  * ifdef HUGE_ADDRBOOKS
- *	It takes up 10 character slots in version 03
+ *	It takes up 10 character slots
  * else
- *	Value restricted to take up 5 character slots in version 02
+ *	Value restricted to take up 5 character slots
  *
  * EntryRef Arr -- An array of N OnDiskEntryRef's, where N is the number
  *                     of entries in the address book.  Each entry is followed
@@ -351,8 +406,9 @@ typedef struct adrbk {
  *
  *      Trailer -- magic number "P # * E @"
  *            <SPACE> N  "\n" the same N as the number of EntryRef entries
- *        (N is 10 if HUGE_ADDRBOOKS, 5 otherwise)
- *           W1 W2 W3 W4 W5 "\n"   These W's are widths for helping
+ *        (N is 10 wide if HUGE_ADDRBOOKS, 5 otherwise)
+ *           DELETED_CNT "\n"  (11 chars wide, count of #DELETED lines)
+ *           W1 W2 W3 W4 W5 W6 W7 "\n"   These W's are widths for helping
  *                             to make a nice display of the addrbook.  Each
  *                             consists of a SPACE followed by a one or two
  *                             digit number.
@@ -363,8 +419,10 @@ typedef struct adrbk {
  *                    W1 = max_nickname_width
  *                    W2 = max_fullname_width
  *                    W3 = max_addrfield_width
- *                    W4 = third_biggest_fullname_width
- *                    W5 = third_biggest_addrfield_width
+ *                    W4 = max_fccfield_width
+ *                    W5 = third_biggest_fullname_width
+ *                    W6 = third_biggest_addrfield_width
+ *                    W7 = third_biggest_fccfield_width
  *
  * Each OnDiskEntryRef looks like:
  *      next_nick  -- 5 bytes, positive integer padded on left with spaces.
@@ -379,9 +437,9 @@ typedef struct adrbk {
 
 #define ADRHASH_FILE_SUFFIX      ".lu"
 #ifdef HUGE_ADDRBOOKS
-#define ADRHASH_FILE_VERSION_NUM "04"
+#define ADRHASH_FILE_VERSION_NUM "10"
 #else
-#define ADRHASH_FILE_VERSION_NUM "03"
+#define ADRHASH_FILE_VERSION_NUM "09"
 #endif
 #define PMAGIC                   "P#*E@"
 #define LEGACY_PMAGIC            "P#*@ "  /* sorry about that */
@@ -398,6 +456,7 @@ typedef struct adrbk {
 #define SIZEOF_UID          SIZEOF_ASCII_LONG
 #define SIZEOF_FILEOFFSET   SIZEOF_ASCII_ULONG
 #define SIZEOF_TIMESTAMP    SIZEOF_ASCII_ULONG
+#define SIZEOF_DELETED_CNT  SIZEOF_ASCII_LONG
 
 #ifdef HUGE_ADDRBOOKS
 
@@ -424,8 +483,10 @@ typedef struct adrbk {
 #define TO_FIND_TIMESTAMP (TO_FIND_SORT_RULE + \
                         (-(SIZEOF_TIMESTAMP + SIZEOF_SPACE)))
 #define TO_FIND_WIDTHS (TO_FIND_TIMESTAMP + \
-		    (-(5 * (SIZEOF_SPACE + SIZEOF_WIDTH) + SIZEOF_NEWLINE)))
-#define TO_FIND_COUNT (TO_FIND_WIDTHS + \
+		    (-(7 * (SIZEOF_SPACE + SIZEOF_WIDTH) + SIZEOF_NEWLINE)))
+#define TO_FIND_DELETED_CNT (TO_FIND_WIDTHS + \
+			(-(SIZEOF_DELETED_CNT + SIZEOF_NEWLINE)))
+#define TO_FIND_COUNT (TO_FIND_DELETED_CNT + \
 			(-(SIZEOF_COUNT + SIZEOF_NEWLINE)))
 #define TO_FIND_TRLR_PMAGIC (TO_FIND_COUNT + \
 			(-(SIZEOF_PMAGIC + SIZEOF_SPACE)))
@@ -439,48 +500,36 @@ typedef struct adrbk {
 #define SIZEOF_HTABLE_ENTRY (SIZEOF_HASH_INDEX + SIZEOF_NEWLINE)
 
 
-#ifdef ANSI
-int             adrbk_add(AdrBk *, char *, char *,
-                          char *, char *, char *, Tag, adrbk_cntr_t *, int *);
-void            adrbk_clearrefs(AdrBk *);
-void            adrbk_close(AdrBk *);
-void            adrbk_partial_close(AdrBk *);
-adrbk_cntr_t    adrbk_count(AdrBk *);
-int             adrbk_delete(AdrBk *, a_c_arg_t);
-char           *adrbk_formatname(char *);
-AdrBk_Entry    *adrbk_get_ae(AdrBk *, a_c_arg_t, Handling);
-int             adrbk_is_in_sort_order(AdrBk *);
-int             adrbk_listadd(AdrBk *, a_c_arg_t, char *);
-int             adrbk_nlistadd(AdrBk *, a_c_arg_t, char **);
-int             adrbk_listdel(AdrBk *, a_c_arg_t, char *);
-AdrBk_Entry    *adrbk_lookup_by_addr(AdrBk *, char *, adrbk_cntr_t *);
-AdrBk_Entry    *adrbk_lookup_by_nick(AdrBk *, char *, adrbk_cntr_t *);
-AdrBk_Entry    *adrbk_newentry(void);
-AdrBk          *adrbk_open(char *, char *, char *, int, int, int);
-long            adrbk_get_nominal_cachesize(AdrBk *);
-long            adrbk_set_nominal_cachesize(AdrBk *, long);
-int             adrbk_sort(AdrBk *, a_c_arg_t, adrbk_cntr_t *, int);
+/*
+ * Prototypes
+ */
+int             adrbk_add PROTO((AdrBk *, a_c_arg_t, char *, char *, char *, \
+				 char *, char *, Tag, adrbk_cntr_t *, int *));
+void            adrbk_clearrefs PROTO((AdrBk *));
+void            adrbk_close PROTO((AdrBk *));
+void            adrbk_partial_close PROTO((AdrBk *));
+adrbk_cntr_t    adrbk_count PROTO((AdrBk *));
+int             adrbk_delete PROTO((AdrBk *, a_c_arg_t, int));
+char           *adrbk_formatname PROTO((char *));
+AdrBk_Entry    *adrbk_get_ae PROTO((AdrBk *, a_c_arg_t, Handling));
+int             adrbk_is_in_sort_order PROTO((AdrBk *, int));
+int             adrbk_listadd PROTO((AdrBk *, a_c_arg_t, char *));
+int             adrbk_nlistadd PROTO((AdrBk *, a_c_arg_t, char **));
+int             adrbk_listdel PROTO((AdrBk *, a_c_arg_t, char *));
+int             adrbk_listdel_all PROTO((AdrBk *, a_c_arg_t));
+AdrBk_Entry    *adrbk_lookup_by_addr PROTO((AdrBk *, char *, adrbk_cntr_t *));
+AdrBk_Entry    *adrbk_lookup_by_nick PROTO((AdrBk *, char *, adrbk_cntr_t *));
+AdrBk_Entry    *adrbk_newentry PROTO((void));
+AdrBk          *adrbk_open PROTO((char *, char *, char *, int, int, int));
+long            adrbk_get_nominal_cachesize PROTO((AdrBk *));
+long            adrbk_set_nominal_cachesize PROTO((AdrBk *, long));
+int             adrbk_sort PROTO((AdrBk *, a_c_arg_t, adrbk_cntr_t *, int));
+void            exp_free PROTO((EXPANDED_S *));
+adrbk_cntr_t	exp_get_next PROTO ((EXPANDED_S **));
+int             exp_is_expanded PROTO((EXPANDED_S *, a_c_arg_t));
+void            exp_set_expanded PROTO((EXPANDED_S *, a_c_arg_t));
+void            exp_unset_expanded PROTO((EXPANDED_S *, a_c_arg_t));
+void            free_ae PROTO((AdrBk *, AdrBk_Entry **));
+char          **parse_addrlist PROTO((char *));
 
-#else /* !ANSI */
-
-int             adrbk_add();
-void            adrbk_clearrefs();
-void            adrbk_close();
-void            adrbk_partial_close();
-adrbk_cntr_t    adrbk_count();
-int             adrbk_delete();
-char           *adrbk_formatname();
-AdrBk_Entry    *adrbk_get_ae();
-int             adrbk_is_in_sort_order();
-int             adrbk_listadd();
-int             adrbk_nlistadd();
-int             adrbk_listdel();
-AdrBk_Entry    *adrbk_lookup_by_addr();
-AdrBk_Entry    *adrbk_lookup_by_nick();
-AdrBk_Entry    *adrbk_newentry();
-AdrBk          *adrbk_open();
-long            adrbk_get_nominal_cachesize();
-long            adrbk_set_nominal_cachesize();
-int             adrbk_sort();
-#endif /* !ANSI */
 #endif /* _ADRBKLIB_INCLUDED */
