@@ -1,5 +1,5 @@
 #if !defined(lint) && !defined(DOS)
-static char rcsid[] = "$Id: imap.c,v 4.49 1994/09/09 17:34:06 mikes Exp $";
+static char rcsid[] = "$Id: imap.c,v 4.79 1996/03/15 07:13:42 hubert Exp $";
 #endif
 /*----------------------------------------------------------------------
 
@@ -16,32 +16,21 @@ static char rcsid[] = "$Id: imap.c,v 4.49 1994/09/09 17:34:06 mikes Exp $";
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
+
+   Pine and Pico are registered trademarks of the University of Washington.
+   No commercial use of these trademarks may be made without prior written
+   permission of the University of Washington.
+
+   Pine, Pico, and Pilot software and its included text are Copyright
+   1989-1996 by the University of Washington.
+
+   The full text of our legal notices is contained in the file called
+   CPYRIGHT, included with this distribution.
+
+
    USENET News reading additions in part by L Lundblade / NorthWestNet, 1993
    lgl@nwnet.net
 
-   Copyright 1989-1994  University of Washington
-
-    Permission to use, copy, modify, and distribute this software and its
-   documentation for any purpose and without fee to the University of
-   Washington is hereby granted, provided that the above copyright notice
-   appears in all copies and that both the above copyright notice and this
-   permission notice appear in supporting documentation, and that the name
-   of the University of Washington not be used in advertising or publicity
-   pertaining to distribution of the software without specific, written
-   prior permission.  This software is made available "as is", and
-   THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
-   WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
-   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
-   NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
-   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
-   LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
-   (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
-   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
-  
-   Pine and Pico are trademarks of the University of Washington.
-   No commercial use of these trademarks may be made without prior
-   written permission of the University of Washington.
-
    Pine is in part based on The Elm Mail System:
     ***********************************************************************
     *  The Elm Mail System  -  Revision: 2.13                             *
@@ -79,23 +68,19 @@ typedef struct mmlogin_s {
 } MMLOGIN_S;
 
 
-#ifdef	ANSI
-int   imap_get_passwd(MMLOGIN_S *, char *, char *, char *);
-void  imap_set_passwd(MMLOGIN_S **, char *, char *, char *);
-#ifdef	DOS
-char  xlate_in(int);
-char  xlate_out(char);
-char *passfile_name(char *, char *);
-int   read_passfile(char *, MMLOGIN_S **);
-void  write_passfile(char *, MMLOGIN_S *);
-int   get_passfile_passwd(char *, char *, char *, char *);
-void  set_passfile_passwd(char *, char *, char *, char *);
-#endif
-
-#else
-int  imap_get_passwd();
-void imap_set_passwd();
-
+/*
+ * Internal prototypes
+ */
+int   imap_get_passwd PROTO((MMLOGIN_S *, char *, char *, char *));
+void  imap_set_passwd PROTO((MMLOGIN_S **, char *, char *, char *));
+#if defined(DOS) || defined(OS2)
+char  xlate_in PROTO((int));
+char  xlate_out PROTO((char));
+char *passfile_name PROTO((char *, char *));
+int   read_passfile PROTO((char *, MMLOGIN_S **));
+void  write_passfile PROTO((char *, MMLOGIN_S *));
+int   get_passfile_passwd PROTO((char *, char *, char *, char *));
+void  set_passfile_passwd PROTO((char *, char *, char *, char *));
 #endif
 
 
@@ -105,9 +90,7 @@ void imap_set_passwd();
  * and to allow mm_searched to return number of matches.
  */
 MAILSTREAM *mm_search_stream;
-MSGNO_S    *mm_search_map	= NULL,
-	   *mm_exclude_map	= NULL;
-int         mm_search_count	= 0L;
+long	    mm_search_count	= 0L;
 
 
 /*
@@ -140,7 +123,16 @@ void
 mm_dlog(string)
     char *string;
 {
-    dprint(0, (debugfile, "IMAP DEBUG: %s\n", string));
+#ifdef	DEBUG
+    time_t      now;
+    struct tm  *tm_now;
+
+    now = time((time_t *)0);
+    tm_now = localtime(&now);
+    dprint(0, (debugfile, "IMAP DEBUG %2.2d:%2.2d:%2.2d %d/%d: %s\n",
+	       tm_now->tm_hour, tm_now->tm_min, tm_now->tm_sec,
+	       tm_now->tm_mon+1, tm_now->tm_mday, string));
+#endif
 }
 
 
@@ -163,27 +155,37 @@ mm_log(string, errflg)
     char        message[300];
     char       *occurance;
     int         was_capitalized;
-    long        now;
+    time_t      now;
     struct tm  *tm_now;
 
-    now = time(0);
+    now = time((time_t *)0);
     tm_now = localtime(&now);
 
-    dprint((errflg == ERROR ? 1 : 2), (debugfile,
-                                     "IMAP %d:%d %d/%d mm_log %s: %s\n",
-                                     tm_now->tm_hour, tm_now->tm_min,
-                                     tm_now->tm_mon+1, tm_now->tm_mday,
-                                     errflg == ERROR ? "ERROR" :
-                                     errflg == WARN  ? "warn"  :
-                                     errflg == PARSE ? "parse" : "babble",
-                                     string));
+    dprint((errflg == ERROR ? 1 : 2),
+	   (debugfile,
+	    "IMAP %2.2d:%2.2d:%2.2d %d/%d mm_log %s: %s\n",
+	    tm_now->tm_hour, tm_now->tm_min, tm_now->tm_sec, tm_now->tm_mon+1,
+	    tm_now->tm_mday,
+	    (errflg == ERROR)
+	      ? "ERROR"
+	      : (errflg == WARN)
+		  ? "warn"
+		  : (errflg == PARSE)
+		      ? "parse"
+		      : "babble",
+	    string));
 
-    /*
-     * don't display if creating new folder
-     * *or* if error on a stream known to be dead
-     */
-    if(ps_global->try_to_create
-       || (ps_global->noticed_dead_stream && !strncmp(string, "[CLOSED]", 8)))
+    if(errflg == ERROR && !strncmp(string, "[TRYCREATE]", 11)){
+	ps_global->try_to_create = 1;
+	return;
+    }
+    else if(ps_global->try_to_create
+       || (ps_global->dead_stream
+	   && (!strncmp(string, "[CLOSED]", 8) || strstr(string, "No-op"))))
+      /*
+       * Don't display if creating new folder OR
+       * warning about a dead stream ...
+       */
       return;
 
     /*---- replace all "mailbox" with "folder" ------*/
@@ -211,8 +213,11 @@ mm_log(string, errflg)
                        (errflg != ERROR && errflg != WARN))
       return; /* Only care about errors; don't print when asked not to */
 
+    if(errflg == ERROR)
+      ps_global->mm_log_error = 1;
+
     /*---- Display the message ------*/
-    q_status_message1(1, errflg == ERROR ? 3 : 2, 5, "\007%s", message);
+    q_status_message(SM_ORDER | SM_DING, 3, 5, message);
     strcpy(ps_global->last_error, message);
 }
 
@@ -236,37 +241,55 @@ mm_notify(stream, string, errflag)
     char       *string;
     long        errflag;
 {
-    int l;
-
     /* be sure to log the message... */
-    sprintf(ps_global->last_error, "%s : %s",
-	    (stream && stream->mailbox) ? stream->mailbox : "-no folder-",
-	    string);
-
-    dprint(1, (debugfile, "IMAP mm_notify %s : %s\n",
+    dprint(1, (debugfile, "IMAP mm_notify %s : %s : %s\n",
                (!errflag) ? "NIL" : 
 		 (errflag == ERROR) ? "error" :
 		   (errflag == WARN) ? "warning" :
 		     (errflag == BYE) ? "bye" : "unknown",
-	       ps_global->last_error));
+	       (stream && stream->mailbox) ? stream->mailbox : "-no folder-",
+	       string));
+
+    sprintf(ps_global->last_error, "%.50s : %.*s",
+	    (stream && stream->mailbox) ? stream->mailbox : "-no folder-",
+	    min(MAX_SCREEN_COLS, 500-70),
+	    string);
+    ps_global->last_error[ps_global->ttyo->screen_cols] = '\0';
 
     /*
      * Then either set special bits in the pine struct or
      * display the message if it's tagged as an "ALERT" or
      * its errflag > NIL (i.e., WARN, or ERROR)
      */
-    if(!strncmp(string, "[TRYCREATE]", 11))
+    if(errflag == BYE){
+	if(stream == ps_global->mail_stream){
+	    if(ps_global->dead_stream)
+	      return;
+	    else
+	      ps_global->dead_stream = 1;
+	}
+	else if(stream && stream == ps_global->inbox_stream){
+	    if(ps_global->dead_inbox)
+	      return;
+	    else
+	      ps_global->dead_inbox = 1;
+	}
+    }
+    else if(!strncmp(string, "[TRYCREATE]", 11))
       ps_global->try_to_create = 1;
-    else if(!strncmp(string, "[READ-ONLY]", l = 11)
-	    || !strncmp(string, "[ALERT]", l = 7))
-      q_status_message2(0, 3, 3, "\007%s : %s",
+    else if(!strncmp(string, "[ALERT]", 7))
+      q_status_message2(SM_MODAL, 3, 3, "%s : %s",
 			(stream && stream->mailbox)
 			  ? stream->mailbox : "-no folder-",
-			string + l);
+			string + 7);
+    else if(!strncmp(string, "[READ-ONLY]", 11))
+      q_status_message2(SM_ORDER | SM_DING, 3, 3, "%s : %s",
+			(stream && stream->mailbox)
+			  ? stream->mailbox : "-no folder-",
+			string + 11);
     else if(errflag && errflag != BYE)
-      q_status_message2(1, (errflag == ERROR) ? 2 : 1, 6, "%s%s",
-			(errflag == ERROR) ? "\007" : "",
-			ps_global->last_error);
+      q_status_message(SM_ORDER | ((errflag == ERROR) ? SM_DING : 0),
+		       3, 6, ps_global->last_error);
 }
 
 
@@ -390,23 +413,8 @@ mm_searched(stream, number)
     MAILSTREAM *stream;
     long        number;
 {
-    if(stream == mm_search_stream){
-	if(mm_search_map){
-	    long i = mn_raw2m(ps_global->msgmap, number);
-	    if(i && !mn_is_cur(mm_search_map, i)
-	       && (!mm_exclude_map || mn_is_cur(mm_exclude_map, i))){
-		mm_search_count++;
-		if(mn_get_cur(mm_search_map) < 1L){
-		    mn_set_cur(mm_search_map, i);
-		}
-		else{
-		    mn_add_cur(mm_search_map, i);
-		}
-	    }
-	}
-	else
-	  mm_search_count++;
-    }
+    if(stream == mm_search_stream)
+      mm_search_count++;
 }
 
 
@@ -429,13 +437,11 @@ mm_login(host, user, passwd, trial)
     long  trial;
 {
     static char junk[] = {'?', '\0'};
-    char      prompt[80], *p;
+    char      prompt[80], *p, *named_user = NULL;
     HelpType  help ;
     int       rc, q_line;
 
-    q_line = -3;				/* 3 from bottom */
-
-    flush_status_messages();
+    q_line = -FOOTER_ROWS(ps_global);			/* 3 from bottom */
 
     if(ps_global->anonymous) {
         /*------ Anonymous login mode --------*/
@@ -450,15 +456,23 @@ mm_login(host, user, passwd, trial)
         return;
     }
 
+    /*
+     * Initialize user name with either /user= value in the stream
+     * being logged into (accessed via the GET_USERNAMEBUF c-client
+     * parameter) or the user name we're running under...
+     */
+    named_user = (char *) mail_parameters(NULL, GET_USERNAMEBUF, NULL);
+    if(trial == 0L)
+      strcpy(user, (named_user && *named_user)
+		     ? named_user : ps_global->VAR_USER_ID);
+
     /*
      * try last working password associated with this host.
-     * NOTE: once we support some sort of user spec (ala {host/user=foo})
-     * we can do a little better cache matching
      */
     if(trial == 0L && imap_get_passwd(mm_login_list, passwd, user, host))
       return;
 
-#ifdef	DOS
+#if defined(DOS) || defined(OS2)
     /* check to see if there's a password left over from last session */
     if(trial == 0L && get_passfile_passwd(ps_global->pinerc,passwd,user,host)){
 	imap_set_passwd(&mm_login_list, passwd, user, host);
@@ -467,24 +481,25 @@ mm_login(host, user, passwd, trial)
 #endif
 
     ps_global->mangled_footer = 1;
-    help = NO_HELP;
-    sprintf(prompt, "HOST: %s  ENTER LOGIN NAME: ", host);
-    strcpy(user, ps_global->VAR_USER_ID);		/* set default */
-    while(1) {
-        rc = optionally_enter(user, q_line, 0, MAILTMPLEN - 1, 1, 0,
-                              prompt, NULL, help, 0);
-        if(rc == 3) {
-            help = help == NO_HELP ? h_oe_login : NO_HELP;
-            continue;
-        }
-        if(rc != 4)
-          break;
-    }
+    if(!(named_user && *named_user)){
+	help = NO_HELP;
+	sprintf(prompt, "HOST: %s  ENTER LOGIN NAME: ", host);
+	while(1) {
+	    rc = optionally_enter(user, q_line, 0, MAILTMPLEN - 1, 1, 0,
+				  prompt, NULL, help, 0);
+	    if(rc == 3) {
+		help = help == NO_HELP ? h_oe_login : NO_HELP;
+		continue;
+	    }
+	    if(rc != 4)
+	      break;
+	}
 
-    if(rc == 1 || !user[0]) {
-        user[0]   = '\0';
-        passwd[0] = '\0';
-        return;
+	if(rc == 1 || !user[0]) {
+	    user[0]   = '\0';
+	    passwd[0] = '\0';
+	    return;
+	}
     }
 
     help = NO_HELP;
@@ -502,14 +517,21 @@ mm_login(host, user, passwd, trial)
     }
 
     if(rc == 1 || !passwd[0]) {
-        strcpy(user,junk);
-        strcpy(passwd, junk);
+	if(!named_user){
+	    strcpy(user,junk);
+	    strcpy(passwd, junk);
+	}
+	else{
+	    user[0]   = '\0';
+	    passwd[0] = '\0';
+	}
+
         return;
     }
 
     /* remember the password for next time */
     imap_set_passwd(&mm_login_list, passwd, user, host);
-#ifdef	DOS
+#if defined(DOS) || defined(OS2)
     /* if requested, remember it on disk for next session */
     set_passfile_passwd(ps_global->pinerc, passwd, user, host);
 #endif
@@ -527,10 +549,10 @@ mm_login(host, user, passwd, trial)
 Result: If error is non serious, the stream is marked as having an error
         and deletes are disallowed until error clears
         If error is serious this goes modal, allowing the user to retry
-        of get a shell escape to fix the condition. When the condition is
+        or get a shell escape to fix the condition. When the condition is
         serious it means that mail existing in the mailbox will be lost
-        if Pine exists without writing so we try to induce the user to 
-        fix the error, go get some one that can fix the error or whatever
+        if Pine exits without writing, so we try to induce the user to 
+        fix the error, go get someone that can fix the error, or whatever
         and don't provide an easy way out.
   ----*/
 long
@@ -539,96 +561,141 @@ mm_diskerror (stream, errcode, serious)
     long        errcode;
     long        serious;
 {
-    int line, ch;
-    char q[200];
+    int  i, j;
+    char *p, *q, *s;
+    static ESCKEY_S de_opts[] = {
+	{'r', 'r', "R", "Retry"},
+	{'f', 'f', "F", "FileBrowser"},
+	{'s', 's', "S", "ShellPrompt"},
+	{-1, 0, NULL, NULL}
+    };
+#define	DE_COLS		(ps_global->ttyo->screen_cols)
+#define	DE_LINE		(ps_global->ttyo->screen_rows - 3)
+
+#define	DE_FOLDER(X)	((X) ? (X)->mailbox : "<no folder>")
+#define	DE_PMT	\
+   "Disk error!  Choose Retry, or the File browser or Shell to clean up: "
+#define	DE_STR1		"SERIOUS DISK ERROR WRITING: \"%s\""
+#define	DE_STR2	\
+   "The reported error number is %s.  The last reported mail error was:"
+    static char *de_msg[] = {
+	"Please try to correct the error preventing Pine from saving your",
+	"mail folder.  For example if the disk is out of space try removing",
+	"unneeded files.  You might also contact your system administrator.",
+	"",
+	"Both Pine's File Browser and an option to enter the system's",
+	"command prompt are offered to aid in fixing the problem.  When",
+	"you believe the problem is resolved, choose the \"Retry\" option.",
+	"Be aware that messages may be lost or this folder left in an",
+	"inaccessible condition if you exit or kill Pine before the problem",
+	"is resolved.",
+	NULL};
+    static char *de_shell_msg[] = {
+	"\n\nPlease attempt to correct the error preventing saving of the",
+	"mail folder.  If you do not know how to correct the problem, contact",
+	"your system administrator.  To return to Pine, type \"exit\".",
+	NULL};
 
     dprint(0, (debugfile,
        "\n***** DISK ERROR on stream %s. Error code %ld. Error is %sserious\n",
-               stream ? stream->mailbox : "<no mail folder>", errcode,
-               serious ? "" : "not "));
+	       DE_FOLDER(stream), errcode, serious ? "" : "not "));
     dprint(0, (debugfile, "***** message: \"%s\"\n\n", ps_global->last_error));
 
     if(!serious) {
-/*        q_status_message2(1, 3, 6,"\007Error saving mail folder \"%s\" : %s",
-                    pretty_fn(stream ? stream->mailbox : ""),
-                          error_description((int)errcode)); */
         if(stream == ps_global->mail_stream) {
             ps_global->io_error_on_stream = 1;
         }
+
         return (1) ;
     }
 
-    line = ps_global->ttyo->screen_rows - 3;
-    MoveCursor(line, 0);
-    CleartoEOLN();
-    StartInverse();
-    Writechar('\007', 0);
-    sprintf(q,  "Serious error saving mail folder \"%s\":",
-                    pretty_fn(stream ? stream->mailbox : ""));
-    PutLine0(line, 0, q);
-    ch = strlen(q); /* Use ch temporarily */
-    MoveCursor(line, ch);
-    while(ch++ < ps_global->ttyo->screen_cols - 1)
-      Writechar(' ', 0);
-/*    sprintf(q, "%-*.*s", ps_global->ttyo->screen_cols - 1,
-            ps_global->ttyo->screen_cols - 1,
-            ps_global->last_error);
-    d_q_status_message();
-    PutLine0(line+1, 0, q); */
-    if(F_ON(F_CAN_SUSPEND,ps_global)) {
-        sprintf(q,
-		"Press return to retry saving or press ^Z to suspend Pine: ");
-    } else {
-        sprintf(q,
-	  "Press return to retry saving or press ! to get to system prompt: ");
-    }
-    PutLine0(line+2, 0, q);
-    Writechar('\007', 0);
-    ch = strlen(q); /* Use ch temporarily */
-    MoveCursor(line+2, ch);
-    while(ch++ < ps_global->ttyo->screen_cols - 1)
-      Writechar(' ', 0);
-    MoveCursor(line+2, strlen(q));
+    while(1){
+	/* replace pine's body display with screen full of explanatory text */
+	ClearLine(2);
+	PutLine1(2, max((DE_COLS - sizeof(DE_STR1)
+					    - strlen(DE_FOLDER(stream)))/2, 0),
+		 DE_STR1, DE_FOLDER(stream));
+	ClearLine(3);
+	PutLine1(3, 4, DE_STR2, long2string(errcode));
+	     
+	PutLine0(4, 0, "       \"");
+	removing_leading_white_space(ps_global->last_error);
+	for(i = 4, p = ps_global->last_error; *p && i < DE_LINE; ){
+	    for(s = NULL, q = p; *q && q - p < DE_COLS - 16; q++)
+	      if(isspace(*q))
+		s = q;
 
-    ch = read_char(0);
-    while(!(ch == F_ON(F_CAN_SUSPEND,ps_global) ? ctrl('Z') : '!'
-	    || ch == '\r')) {
-        Writechar('\007', 0);
-        ch = read_char(0);
-    }
-    switch(ch) {
-      case '\r':
-        Write_to_screen("Retry");
-        EndInverse();
-        MoveCursor(line -2, 0);
-        fflush(stdout);
-        return (0L);
-       
-      case '!':
-        EndInverse();
-        end_keyboard(ps_global != NULL ? F_ON(F_USE_FK,ps_global) : 0);
-        end_tty_driver(ps_global);
-        printf("\n\nPlease attempt to correct the error preventing the saving of the mail folder.\n");
-        printf("For example if the disk is out of space try removing uneeded files.\n");
-        printf("You might also contact your system administrator.\n");
-        printf("Give the command \"exit\" to get back to Pine and try saving the folder again.\n");
+	    if(*q && s)
+	      q = s;
 
+	    while(p < q)
+	      Writechar(*p++, 0);
+
+	    if(*(p = q)){
+		ClearLine(++i);
+		PutLine0(i, 0, "        ");
+		while(*p && isspace(*p))
+		  p++;
+	    }
+	    else{
+		Writechar('\"', 0);
+		CleartoEOLN();
+		break;
+	    }
+	}
+
+	ClearLine(++i);
+	for(j = ++i; i < DE_LINE && de_msg[i-j]; i++){
+	    ClearLine(i);
+	    PutLine0(i, 0, "  ");
+	    Write_to_screen(de_msg[i-j]);
+	}
+
+	while(i < DE_LINE)
+	  ClearLine(i++);
+
+	switch(radio_buttons(DE_PMT, -FOOTER_ROWS(ps_global), de_opts,
+			     'r', 0, NO_HELP, RB_FLUSH_IN | RB_NO_NEWMAIL)){
+	  case 'r' :				/* Retry! */
+	    ps_global->mangled_screen = 1;
+	    return(0L);
+
+	  case 'f' :				/* File Browser */
+	    {
+		char full_filename[MAXPATH+1], filename[MAXPATH+1];
+
+		filename[0] = '\0';
+		build_path(full_filename, ps_global->home_dir, filename);
+		file_lister("DISK ERROR", full_filename, filename, FALSE,
+			    FB_SAVE);
+	    }
+
+	    break;
+
+	  case 's' :
+	    EndInverse();
+	    end_keyboard(ps_global ? F_ON(F_USE_FK,ps_global) : 0);
+	    end_tty_driver(ps_global);
+	    for(i = 0; de_shell_msg[i]; i++)
+	      puts(de_shell_msg[i]);
+
+	    /*
+	     * Don't use our piping mechanism to spawn a subshell here
+	     * since it will the server (thus reentering c-client).
+	     * Bad thing to do.
+	     */
 #ifdef	_WINDOWS
 #else
-        system("csh");
+	    system("csh");
 #endif
+	    init_tty_driver(ps_global);
+	    init_keyboard(F_ON(F_USE_FK,ps_global));
+	    break;
+	}
 
-        init_tty_driver(ps_global);
-        init_keyboard(F_ON(F_USE_FK,ps_global));
-        MoveCursor(line - 2, 0);
-        return (0L); /* go retry it again */
-       
-      case ctrl('Z'):
-        EndInverse();
-        do_suspend(ps_global);
-        break;
+	if(ps_global->redrawer)
+	  (*ps_global->redrawer)();
     }
-    return(0);
 }
 
 
@@ -676,6 +743,8 @@ void
 context_bboard(name)
     char *name;
 {
+    dprint(4, (debugfile, "====== context_bboard: (%s)\n", name));
+
 #ifdef NEWBB
     /*----
 
@@ -688,13 +757,14 @@ context_bboard(name)
     }
 #endif
 
+    if(!ps_global->show_dot_names && *name == '.')
+       return;
+
     find_folder_count++;
     if(find_folder_list){
 	FOLDER_S *new_f = new_folder(name);
 	folder_insert(folder_total(find_folder_list), new_f, find_folder_list);
     }
-
-    dprint(4, (debugfile, "====== context_bboard: (%s)\n", name));
 }
 
 
@@ -731,31 +801,85 @@ mm_flags(stream,number)
 }
 
 
+/*
+ * Seconds afterwhich it's ok to prompt for connention severage...
+ */
+#define	TO_BAIL_THRESHOLD	60L
+
+
+/*
+ * pine_tcptimeout - C-client callback to handle tcp-related timeouts.
+ */
+long
+pine_tcptimeout(t)
+    long t;
+{
+    long rv = 1L;			/* keep trying by default */
+    int  ch;
+
+    if(ps_global->noshow_timeout)
+      return(rv);
+
+    /*
+     * Prompt after a minute (since by then things are probably really bad)
+     * A prompt timeout means "keep trying"...
+     */
+    if(t >= TO_BAIL_THRESHOLD){
+	int clear_inverse;
+
+	ClearLine(ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global));
+	if(clear_inverse = !InverseState())
+	  StartInverse();
+
+	PutLine1(ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global), 0,
+       "\007Waited %s seconds for server reply.  Break connection to server? ",
+	   long2string(t));		/* if not DOS, add fflush() */
+	CleartoEOLN();
+	fflush(stdout);
+	flush_input();
+	ch = read_char(7);
+	if(ch == 'y' || ch == 'Y')
+	  rv = 0L;
+
+	if(clear_inverse)
+	  EndInverse();
+
+	ClearLine(ps_global->ttyo->screen_rows - FOOTER_ROWS(ps_global));
+    }
+
+    if(rv == 1L){			/* just warn 'em something's up */
+	q_status_message1(SM_ORDER, 0, 0,
+		  "Waited %s seconds for server reply.  Still Waiting...",
+		  long2string(t));
+	flush_status_messages(0);	/* make sure it's seen */
+    }
+
+    mark_status_dirty();		/* make sure it get's cleared */
+    return(rv);
+}
+
+
 
 /*----------------------------------------------------------------------
    Exported method to retrieve logged in user name associated with stream
 
-   Args: mbox -- mailbox to find associated login name with.
+   Args: host -- host to find associated login name with.
 
  Result: 
   ----*/
 char *
-imap_user(mbox)
-    char *mbox;
+cached_user_name(host)
+    char *host;
 {
     MMLOGIN_S *l;
-    NETMBX     netmbox;
 
-    if((l = mm_login_list) && mbox
-       && mail_valid_net_parse(mbox, &netmbox) && !netmbox.anoflag){
-	do{
-	    if(!strucmp(netmbox.host, l->host))
-	      break;
-	}
-	while(l = l->next);
-    }
+    if((l = mm_login_list) && host)
+      do
+	if(!strucmp(host, l->host))
+	  return(l->user);
+      while(l = l->next);
 
-    return(l ? l->user : NULL);
+    return(NULL);
 }
 
 
@@ -765,7 +889,8 @@ imap_get_passwd(l, passwd, user, host)
     char *passwd, *user, *host;
 {
     while(l){
-	if(!strucmp(host, l->host)){
+	/* host names and, if present, user name, must match */
+	if(!strucmp(host, l->host) && (!*user || !strcmp(user, l->user))){
 	    strcpy(user, l->user);
 	    strcpy(passwd, l->passwd);
 	    break;
@@ -831,7 +956,7 @@ imap_flush_passwd_cache()
  * DOS-specific functions to support caching username/passwd/host
  * triples on disk for use from one session to the next...
  */
-#ifdef	DOS
+#if defined(DOS) || defined(OS2)
 
 #define	FIRSTCH		0x20
 #define	LASTCH		0x7e
