@@ -1,5 +1,5 @@
 /*----------------------------------------------------------------------
-  $Id: os-nxt.h,v 5.17 1994/08/23 02:47:02 hubert Exp $
+  $Id: os-nxt.h,v 5.38 1996/03/15 06:43:47 hubert Exp $
 
             T H E    P I N E    M A I L   S Y S T E M
 
@@ -14,28 +14,17 @@
 
    Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 
-   Copyright 1989-1994  University of Washington
 
-    Permission to use, copy, modify, and distribute this software and its
-   documentation for any purpose and without fee to the University of
-   Washington is hereby granted, provided that the above copyright notice
-   appears in all copies and that both the above copyright notice and this
-   permission notice appear in supporting documentation, and that the name
-   of the University of Washington not be used in advertising or publicity
-   pertaining to distribution of the software without specific, written
-   prior permission.  This software is made available "as is", and
-   THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
-   WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
-   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
-   NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
-   INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
-   LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
-   (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
-   WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
-  
-   Pine and Pico are trademarks of the University of Washington.
-   No commercial use of these trademarks may be made without prior
-   written permission of the University of Washington.
+   Pine and Pico are registered trademarks of the University of Washington.
+   No commercial use of these trademarks may be made without prior written
+   permission of the University of Washington.
+
+   Pine, Pico, and Pilot software and its included text are Copyright
+   1989-1996 by the University of Washington.
+
+   The full text of our legal notices is contained in the file called
+   CPYRIGHT, included with this distribution.
+
 
    Pine is in part based on The Elm Mail System:
     ***********************************************************************
@@ -91,8 +80,8 @@
 
 
 /*----------------------------------------------------------------------
-    Turn this on if it's OK to encode sent message text containing
-  "From " at the beginning of any line.
+    Turn this on to trigger QP encoding of sent message text if it contains
+  "From " at the beginning of a line or "." on a line by itself.
  ----*/
 /* #define ENCODE_FROMS */
 
@@ -101,9 +90,11 @@
 /*----------------------------------------------------------------------
     Timeouts (seconds)
  ----*/
-#define NEW_MAIL_TIME      (150) /* How often to check for new mail. There's 
-                                   some expense in doing this so it shouldn't
-                                   be done too frequently.  */
+#define DF_MAILCHECK      "150" /* How often to check for new mail, by
+				   default.  There's some expense in doing
+				   this so it shouldn't be done too
+				   frequently.  (Can be set in config
+				   file now.)  */
 
 /*----------------------------------------------------------------------
     Check pointing (seconds)
@@ -127,7 +118,17 @@
  In scrolling through text, the number of lines from the previous
  screen to overlap when showing the next screen.  Usually set to two.
  ----*/
-#define	OVERLAP		(2) 
+#define	DF_OVERLAP	"2"
+
+
+
+/*----------------------------------------------------------------------
+ Default fill column for pine composer and maximum fill column.  The max
+ is used to stop people from setting their custom fill column higher than
+ that number.  Note that DF_FILLCOL is a string but MAX_FILLCOL is an integer.
+ ----*/
+#define	DF_FILLCOL	"74"
+#define	MAX_FILLCOL	80
 
 
 
@@ -146,6 +147,7 @@
 #define POSTPONED_MAIL            "postponed-mail"
 #define POSTPONED_MSGS            "postponed-msgs"
 #define INTERRUPTED_MAIL          ".pine-interrupted-mail"
+#define DEADLETTER                "dead.letter"
 #define DF_MAIL_DIRECTORY         "mail"
 #define INBOX_NAME                "INBOX"
 #define DF_SIGNATURE_FILE         ".signature"
@@ -164,6 +166,13 @@
 #define DF_ADDRESSBOOK            ".addressbook"
 #define DF_BUGS_FULLNAME          "Pine Developers"
 #define DF_BUGS_ADDRESS           "pine-bugs@cac.washington.edu"
+#define DF_SUGGEST_FULLNAME       "Pine Developers"
+#define DF_SUGGEST_ADDRESS        "pine-suggestions@cac.washington.edu"
+#define DF_PINEINFO_FULLNAME      "Pine-Info News Group"
+#define DF_PINEINFO_ADDRESS       "pine-info@cac.washington.edu"
+#define DF_LOCAL_FULLNAME         "Local Support"
+#define DF_LOCAL_ADDRESS          "postmaster"
+#define DF_KBLOCK_PASSWD_COUNT    "1"
 
 /*----------------------------------------------------------------------
    The default printer when pine starts up for the first time with no printer
@@ -190,23 +199,6 @@
 /*------ If our compiler doesn't understand type void ------------------*/
 /* #define void char     /* no void in compiler */
 
-/******* These are included in c-client/osdep.h. ************************/
-#ifdef notdef
-/*-------- Standard ANSI functions usually defined in stdlib.h ---------*/
-#include <stdlib.h>  /* declarations of many common functions */
-/*---- Declare getenv() if not already declared ------------------------*/
-/* char *getenv();      /* often declared in stdlib.h */
-
-
-/*------- Standard string functions ------------------------------------*/
-#include <string.h>
-/*---- Declare these if not already declared ---------------------------*/
-/* char *strcpy(), *strncpy();  /* usually in string.h */
-
-
-/*------- Include types.h if you have it -------------------------------*/
-#include <sys/types.h>
-#endif /*****************************************************************/
 
 
 /*------- Some more includes that should usually be correct ------------*/
@@ -226,26 +218,38 @@
 
 
 /*--------------- signal.h ---------------------------------------------*/
-/* #include <signal.h>      /* sometimes both required, sometimes */
-#include <sys/signal.h>  /* only one or the other */
+#include <signal.h>      /* sometimes both required, sometimes */
+/* #include <sys/signal.h>  /* only one or the other */
 
-#define SigType void     /* value returned by signal() is void * */
-/* #define SigType int      /* value returned by signal() is int * */
+#define SigType void     /* value returned by sig handlers is void */
+/* #define SigType int      /* value returned by sig handlers is int */
 
 /* #define POSIX_SIGNALS    /* use POSIX signal semantics (ttyin.c) */
 
-/*--- For AT&T-derviced systems ----------------------------------------*/
+/*--- For AT&T-derived systems ----------------------------------------*/
 /* #define signal(a,b) sigset(a, b) /* sigset is just like signal */
 /* extern void (*sigset())();       /* declare sigset if not declared already */
 
 
 
+/*-------------- A couple typedef's for integer sizes ------------------*/
+typedef unsigned int usign32_t;
+typedef unsigned short usign16_t;
+
+
+
 /*-------------- qsort argument type -----------------------------------*/
 #define QSType void  /* qsort arg is of type void * */
 /* #define QSType char  /* qsort arg is of type char * */
 
 
 
+/*-------------- fcntl flag to set non-blocking IO ---------------------*/
+/*#define	NON_BLOCKING_IO	O_NONBLOCK		/* POSIX style */
+#define	NON_BLOCKING_IO	FNDELAY		/* good ol' bsd style  */
+
+
+
 /*------ how help text is referenced (always char ** on Unix) ----------*/
 #define HelpType char **
 #define NO_HELP (char **)NULL
@@ -266,6 +270,7 @@
 #define HAVE_WAIT_UNION  /* the arg to wait is a union wait * */
 
 
+
 /*-------- Is window resizing available? -------------------------------*/
 #define RESIZING  /* SIGWINCH and friends */
 
@@ -278,10 +283,18 @@
 
 /*----- The usual sendmail configuration for sending mail on Unix ------*/
 #define SENDMAIL	"/usr/lib/sendmail"
-#define SENDMAILFLAGS	"-oi -oem -t"	/* ignore dots in incoming mail,
-					   mail back errors, and
-					   get recipients from To, cc,
-					   and bcc header lines.      */
+#define SENDMAILFLAGS	"-bs -odb -oem"	/* send via smtp with backgroud
+					   delivery and mail back errors */
+
+
+/*----------------------------------------------------------------------
+   If no nntp-servers are defined, this program will be used to post news.
+ ----*/
+/* #define SENDNEWS	"/usr/local/bin/inews -h"	/* news posting cmd */
+
+
+/*--------- Program employed by users to change their password ---------*/
+#define	PASSWD_PROG	"/bin/passwd"
 
 
 /*-------------- A couple constants used to size arrays ----------------*/
